from dateroll.settings import settings
from dateroll.date.date import Date
from dateroll.ddh.ddh import calmath, cals, ddh
from dateroll.duration.duration import Duration
from dateroll.schedule.schedule import Schedule

# for backwards compatibility testing
# import sys
# import traceback
# try:
#     from dateroll.ddh.ddh import calmath, cals, ddh
#     from dateroll.date.date import Date
#     from dateroll.duration.duration import Duration
#     from dateroll.schedule.schedule import Schedule
#     ddh("t+3m1bd|NYuLN")
#     print("asdfasdf", sys.version.split(" ")[0], "✅" * 30)
# except Exception as e:
#     traceback.print_exc()
#     print("asdfasdf", sys.version.split(" ")[0], "❌" * 30, str(e))



import datetime
import copy
import math
import warnings
import dateutil
import dateutil.relativedelta
import numpy as np

from dateroll.calendars.calendarmath import calmath
import dateroll.parser.parsers as parsersModule
import calendar

# NOTE: functools cache is defined starting from python 3.9
# older versions of python can use functools lru_cache(maxsize=None)
# for now, leave this commented out because it is not used
#
## from functools import cache
## import code

from dateroll.settings import settings

from dateroll.utils import color, xprint, add_none, combine_none

cals = calmath.cals

period_order = (*"yhsqmwd", "cals", "modifier")

DurationLike = (dateutil.relativedelta, datetime.timedelta)
WEEKEND_CALENDAR = "WE"

APPROX = {
    "1y1d": 14610 / 40,  # exact average length for post 1582AD change
    "1bd1d": 14610 / 40 / 252,  # assuming 252 denominator
}


class Duration(dateutil.relativedelta.relativedelta):
    global DEBUG_PRINT
    """
    inherits from relativedelta

    Duration class represents a distance between two dates. The two dates MAY or MAY NOT be known.

    3 modes of operation:

        1 - Distance ignoring holidays
            e.g. 3 months from today, or t+3m
                why? normal parlance: 3 months from now
        2 - Distance of only non-holidays
            e.g. 2 business days from today (skip weekends and federal holidays), or t+2bd|WEuNY
                why? payment adjustment; bond settlement is t+2bd|WEuFED
        3 - Distance ignoring holidays + distance of only non-holidays
            e.g. 1 months from today and then 2 business days, or t+1m2bd|WEuNY
                why? payment adjustment: credit card is due on the 2nd business day in one month

    A duration can have a number of:
        business days
        days (1d = 1.44bd)
        weeks  (1w=7d)
        months (1m approx 28-31d)
        years (1y=12m)

    ^ in order of seniority


    We define the "reasonable business period" (naiively) as the interval of (t-100y,t+100y)
        this could be extended: from 1582 AD to future if we have historical vectors for holiday calendar changes (i.e. juneteenth 1st occured on X, to map backwards math, and assume last forward for futuree math (e.g. last jubilee was 5cd, and happens every 50years..))
        realistically 5-7 year perfect lookup would be more than substantial for more use cases.
    """

    def __init__(
        self,
        # switch mdy to days, months, years to match td and rd
        # primary initialization properties (match dateutil.relativedelta)
        years=0,
        months=0,
        # auxlillary initializers
        y=0,
        Y=0,
        year=0,
        q=0,
        Q=0,
        quarter=0,
        quarters=0,
        m=0,
        M=0,
        month=0,
        weeks=0,
        days=0,
        w=0,
        W=0,
        week=0,
        d=0,
        D=0,
        day=0,
        # bd initializer, must be none as 0 and no zero are different
        bd=None,
        # axuliliary bd initi
        BD=None,
        # bd adjustment initializer
        cals=None,
        modified=False,
        # anchor dates (if duration is the result of math with some date)
        _anchor_start=None,
        _anchor_end=None,
        _anchor_months=None,
        _anchor_days=None,
        # debug
        debug=False,
    ):
        """
        y = year
        q = quarter
        m = month
        w = week
        d = day
        bd = business days
        cals = list of 2-letter codes for calendars
        modified = False or True (True means stay in month with BD adjustment)
        """
        # merge y/m/w/d
        self.years = years + year + y + Y
        self.months = months + month + m + M
        self.days = days + day + d + D

        # collapse quarters into months
        self.months += 3 * (quarters + quarter + Q + q)

        # collapse weeks into days
        self.days += 7 * (weeks + week + W + w)

        # assign modified
        self.modified = modified

        # roll up months>12 into years
        if abs(self.months) >= 12:
            self.years += int(self.months / 12)
            self.months -= 12 * int(self.months / 12)

        # process anchor periods / dates
        self.process_anchor_dates(_anchor_start, _anchor_end)

        # merge bd - FLOAT(x) ensures sign is copied for direction of BD travel
        if bd is not None and BD is not None:
            self.bd = bd + BD
        elif bd is not None:
            self.bd = float(bd)
        elif BD is not None:
            self.bd = float(BD)
        else:
            if cals is not None:
                # implicity calendar for weekends will be set in validate_cals
                self.bd = float(0.0)
            else:
                # definately no bd adjustment
                self.bd = None

        # valid cals
        self._validate_cals(cals)

        # debug
        self.debug = debug

    def process_anchor_dates(self, _anchor_start, _anchor_end):
        """
        if a period is the result of dates, we can be "more exact" when the user needs the number of days in the period
        """

        self._anchor_start = _anchor_start
        self._anchor_end = _anchor_end

        from dateroll.date.date import Date

        if self._anchor_start and self._anchor_end:
            # anchor months -- month diff without days and years collapses into months
            self._anchor_start = Date.from_datetime(self._anchor_start)
            self._anchor_end = Date.from_datetime(self._anchor_end)
            ydiff = self._anchor_end.year - self._anchor_start.year
            mdiff = self._anchor_end.month - self._anchor_start.month
            diff = mdiff + ydiff * 12
            self_anchor_years = ydiff
            self._anchor_months = diff

            # put anchor dates as a total days without subtracting month years
            self._anchor_days = (self._anchor_end.date - self._anchor_start.date).days

    @staticmethod
    def from_string(string):
        letters = [chr(i) for i in range(65, 65 + 26)]

        def gen():
            yield letters.pop(0)

        if isinstance(string, str):
            durs, s = parsersModule.parseDurationString(string, gen)
            # first generating string is always A because of alphabetical order
            if len(durs) == 1 and s in ("+A", "A"):

                return durs["A"]
            else:
                raise TypeError(f"Could not validate duration string: {string}")
        else:
            raise TypeError(f"Must be string not {type(string).__name__}")

    @staticmethod
    def from_relativedelta(rd, _anchor_start=None, _anchor_end=None):
        if isinstance(rd, Duration):
            return rd
        elif isinstance(rd, dateutil.relativedelta.relativedelta):
            return Duration(
                years=rd.years,
                months=rd.months,
                days=rd.days,
                _anchor_start=_anchor_start,
                _anchor_end=_anchor_end,
            )
        elif isinstance(rd, datetime.timedelta):
            return Duration.from_timedelta(
                rd, _anchor_start=_anchor_start, _anchor_end=_anchor_end
            )
        else:
            raise TypeError(
                f"Must be relativedelta (or timedelta) not {type(rd).__name__}"
            )

    @staticmethod
    def from_timedelta(td, _anchor_start=None, _anchor_end=None):
        if isinstance(td, Duration):
            return td
        elif isinstance(td, datetime.timedelta):
            return Duration(
                days=td.days, _anchor_start=_anchor_start, _anchor_end=_anchor_end
            )
        else:
            raise TypeError(f"Must be timedelta not {type(td).__name__}")

    def _validate_cals(self, cals):
        """
        validates calendars are correct
        """
        _cals = set()
        if cals is not None:

            # str parser if required
            if isinstance(cals, str):
                cals = parsersModule.parseCalendarUnionString(cals)

            # process normally
            if isinstance(cals, (list, set, tuple)):
                for cal in cals:
                    if isinstance(cal, str):
                        if len(cal) in (2, 3):
                            if cal in calmath.cals:
                                _cals |= {
                                    cal,
                                }
                            else:
                                raise ValueError(f"Calendar {cal} not found")
                        else:
                            raise Exception(
                                f"Calendars must be 2 or 3 letter strings (not {cal})"
                            )
                    else:
                        raise Exception(
                            f"Calendars must be strings (not {type(cal).__name__})"
                        )
            # implicit weekend when using cals
            _cals |= set([WEEKEND_CALENDAR])
            _cals = tuple(sorted(_cals))

        else:
            if self.bd is not None:
                # implicit weekend calendar if bd is defined
                _cals = (WEEKEND_CALENDAR,)
            else:
                _cals = None

        self.cals = _cals

    def __eq__(self, o):
        """
        equality
        """
        # convert td to rd

        if isinstance(o, int):
            return self.just_exact_days == o

        if isinstance(o, datetime.timedelta):
            o = dateutil.relativedelta.relativedelta(days=o.days)

        if isinstance(o, (dateutil.relativedelta.relativedelta, Duration)):
            if self.years == o.years:
                if self.months == o.months:
                    if self.days == o.days:
                        if isinstance(o, Duration):
                            if self.bd == o.bd:
                                if self.cals == o.cals:
                                    if self.modified == o.modified:
                                        if (
                                            self._anchor_end == o._anchor_end
                                            and self._anchor_start == o._anchor_start
                                        ):
                                            return True
                        else:
                            return True

            if isinstance(o, Duration):
                """
                for durations from date subtraction there is equivalency to whole units
                4/15/24-1/15/23 == 91d == 1y3m == 14m, and all combinations thereof, and vice-versa
                """

                if self.compare_anchors(self, o) or self.compare_anchors(o, self):
                    return True

            return False
        else:
            raise TypeError(f"Cannot compare with {type(o).__name__}")

    def compare_anchors(self, a, b):
        if hasattr(a, "_anchor_months"):
            am = a._anchor_months
            ad = a._anchor_days
            cond1 = am == b.months + b.years * 12
            cond2 = ad == b.days
            if cond1 and cond2:
                return True
        return False

    @property
    def relativedelta(self):
        """
        create relativedelta
        """
        rd = dateutil.relativedelta.relativedelta(
            years=self.years, months=self.months, days=self.days
        )

        return rd

    @property
    def just_days(self):
        # count cal days approx warn return always
        return self._just_days()

    @property
    def just_approx_days(self):
        #  always approx even if perfect days
        return self._just_days(_force_approx=True)

    @property
    def just_exact_days(self):
        # raise execption instead of warn
        return self._just_days(_force_exact=True)
    
    @property
    def just_bds(self):
        if self._anchor_start and self._anchor_end:
            return calmath.diff(self._anchor_start,self._anchor_end,self.cals,settings.ie)
        else:
            raise Exception('Cannot compute BDs on a duration that did not result from two dates')


    def _just_days(self, _force_approx=False, _force_exact=False):
        """
        if Duration is anchored (it's the result of date math)
            We know the EXACT days implied, so return the count

        If not,
            We MAY or MAY NOT be able to know the exact days

            years, months, and non-None BD's trigger approx calcs, and a WARNING is issued
            pure days and weeks is exact always
        """
        warns = []
        just_days = 0

        if self._anchor_start and self._anchor_end and not _force_approx:
            # i am anchored
            just_days += (self._anchor_end - self._anchor_start)._anchor_days

        else:
            # i am not anchored
            if self.years != 0:
                yeardays = APPROX["1y1d"] * self.years
                just_days += yeardays
                warns.append(f"≈{yeardays:.6f}d")

            if self.months != 0:
                monthdays = APPROX["1y1d"] / 12 * self.months
                just_days += monthdays
                warns.append(f"≈{monthdays:.6f}d")

            just_days += self.days

            if self.bd is not None:
                if self.bd != 0:
                    dbds = APPROX["1bd1d"] * self.bd
                    just_days += dbds
                    warns.append(f"≈{dbds:.6f}d")

        if len(warns) > 0:
            w = ",".join(warns)
            message = f"[dateroll] just_days using approx: {w} "
            if not _force_exact:
                warnings.warn(message)
            else:
                raise ValueError(message)

        return just_days

    def math(self, b, direction):
        """
        c = a + b
        """
        from dateroll.date.date import DateLike, Date

        a = self
        # duration + duration
        if isinstance(b, Duration):
            """
            combine both

            3 potential sign changers
            if a has direction use as multiplier
            if b has direciton use as multiple
            if subtraction use incoming direction as multipler
            """

            # non bd adjustments
            years = a.years + direction * b.years
            months = a.months + direction * b.months
            days = a.days + direction * b.days

            if np.sign(years) != np.sign(months):
                # we adjust it to have same signs, since we dont have anchor date we can not adjust the date
                rd_adjusted = dateutil.relativedelta.relativedelta(
                    months=years * 12 + months
                )
                years = rd_adjusted.years
                months = rd_adjusted.months

            # bd adjustments w/o MOD
            bd = add_none(a.bd, b.bd, direction)  # none or add
            cals = combine_none(a.cals, b.cals)  # union

            # modified / if either has, inherit it
            modified = a.modified or b.modified

            # check combined Duration net direction
            dur = Duration(
                years=years,
                months=months,
                days=days,
                bd=bd,
                cals=cals,
                modified=modified,
            )

            return dur

        elif isinstance(b, dateutil.relativedelta.relativedelta):
            """
            combine rd with self via copy then combine years+months+days from rd into self
            """
            dur = copy.deepcopy(self)
            dur.years += b.years
            dur.months += b.months
            dur.days += b.days

            return dur

        elif isinstance(b, DateLike):
            """
            if direction > 0:
                    (Date + Duration)
                or  (Duration + Date)
            if direction < 0:
                        (Date - Duration)
            NOT (Duration - Date)
                Date.__rsub__ throws TypeError before it gets here (non-sensical situation)
            """

            modifed = self.adjust_from_date(b, direction)

            dt = Date.from_datetime(modifed)
            return dt

        else:
            raise NotImplementedError

    def adjust_from_date(self, date_unadj, direction):
        """
        4 steps:
            1 - if being subtracted, negate
            2 - perform non-holiday adjustments first, i.e. y's, m's, d's (returns datetime.date)
            3 - perform holiday adjustments second, i.e. bd's (impliying roll from direction via sign on neg, includes -0.0 as backwards)
            4 - perform modified operation, if supplied
            5 - cast datetime.date back into Date
        """
        (
            xprint(f"Adjusting date [[{date_unadj}]] with [[{direction}*{self}]]")
            if self.debug
            else None
        )

        # 1 negate if being subtracted
        if direction < 0:
            # Date - Duration
            # if negative, flip sign using __neg__
            negated_self = self.__neg__()
            (
                xprint("negation", "before:    ", self, "after:    ", negated_self)
                if self.debug
                else None
            )
        else:
            negated_self = self
            xprint("negation: none") if self.debug else None

        # 2 non-holiday adjustments, add D,M,Y
        date_nonhol_adj = date_unadj.date + negated_self.relativedelta
        (
            xprint(lbl="cal adj", before=date_unadj, after=date_nonhol_adj)
            if self.debug
            else None
        )

        # 3 holiday adjustment, add BD and if modified check, if BD results in diff month, bounce
        if negated_self.bd is not None:
            bd_sign, date_hol_adj = negated_self.adjust_bds(date_nonhol_adj)
            (
                xprint(lbl="hol adj", before=date_nonhol_adj, after=date_hol_adj)
                if self.debug
                else None
            )
            if negated_self.modified:
                # modifying from non-hol adjusted to hol-ajusted, could be subjective arg in future that modification is from date_unadj to date_hol_adj
                date_modifed = negated_self.apply_modifier(
                    date_nonhol_adj, date_hol_adj, bd_sign
                )
                (
                    xprint(lbl="mod adj", before=date_hol_adj, after=date_modifed)
                    if self.debug
                    else None
                )
            else:
                date_modifed = date_hol_adj
                xprint("mod: no mod") if self.debug else None
        else:
            date_modifed = date_nonhol_adj
            xprint("no modifier") if self.debug else None

        # 4 convert back to Date
        from dateroll.date.date import Date

        Date_modified = Date.from_datetime(date_modifed)
        Date_modified.origin_dur_date = date_unadj
        Date_modified.origin_dur_cals = self.cals
        Date_modified.origin_dur_modified = self.modified

        return Date_modified

    def adjust_bds(self, from_date):
        """
        This moves business days, using the sign of BD, yes, -0.0 means go backwards to the previous business day
        note: calmath.add_bd ALSO handles -0.0BD as well

        returns TUPLE, the holiday/bizday ajusted date, and it's sign: (bd_sign,date_hol_adj)
        """

        if self.bd is None:
            # non-case
            return (1, from_date)

        bd_sign = math.copysign(1, self.bd)
        if bd_sign >= 0:
            # FOLLOWING ROLL CONVENTION is handled in +0.0 BD
            date_hol_adj = calmath.add_bd(from_date, abs(self.bd), cals=self.cals)
        else:
            # PREVIOUS ROLL CONVENTION is handled in -0.0 BD
            date_hol_adj = calmath.sub_bd(from_date, abs(self.bd), cals=self.cals)

        return bd_sign, date_hol_adj

    def __hash__(self):
        return hash(str(self.__dict__))

    def apply_modifier(self, before, after, bd_sign):
        """
        based upon the sign of the business date (yes, includes -0.0 as a previous) perform MODIFIER

        FOLLOWING is handled in +0.0 BD
        PREVIOUS is handled in -0.0 BD

        MODOFIED FOLLOWING is +0.0, then HERE
        MODIFIED PREVIOUS is -0.0, then HERE

        """
        from dateroll.date.date import Date

        if after.month != before.month or after.year != before.year:
            if bd_sign > 0:
                # get the latest calendar date of before
                _, num_days = calendar.monthrange(before.year, before.month)

                after = Date(before.year, before.month, num_days)
                after = after - Duration(bd=0, cals=self.cals)

            else:
                after = Date(before.year, before.month, 1)
                after = after + Duration(bd=0, cals=self.cals)

        return after

    def __radd__(self, x):
        xprint(type(x), "radd") if self.debug else None
        return self.math(x, 1)

    def __rsub__(self, x):
        xprint(type(x), "rsub") if self.debug else None

        return self.math(x, -1)

    def __add__(self, x):
        xprint(type(x), "add") if self.debug else None
        return self.math(x, 1)

    def __sub__(self, x):
        xprint(type(x), "sub") if self.debug else None
        from dateroll import Date

        if isinstance(x, Date):
            raise TypeError("Cannot sub Date from Duration")
        return self.math(x, -1)

    def __iadd__(self, x):
        xprint(type(x), "iadd", type(self.math(x, 1))) if self.debug else None
        return self.math(x, 1)

    def __isub__(self, x):
        xprint(type(x), "isub") if self.debug else None
        return self.math(x, -1)

    def __neg__(self):
        """
        apply negative across all units (distributive property)
        """
        copy_self = copy.deepcopy(self)
        copy_self.years = -1 * self.years
        copy_self.months = -1 * self.months
        copy_self.days = -1 * self.days

        if copy_self.bd is not None:
            copy_self.bd = -1 * float(self.bd)

        return copy_self

    def __pos__(self):
        return self

    def __repr__(self):
        """
        repr sorts units by seniority specificed in global

        future note: implicity call the simplify() method before sorting, and repr on the simpllifed version, not direct on __dict__
        consider moving ny/nm/nd/nbd to dict on class??
        """
        d = self.__dict__
        items = {k: d[k] for k in period_order if k in d}
        constructor = ""
        for k, v in self.__dict__.items():
            if v != None and k != "debug" and not k.startswith("_"):
                if k == "cals":
                    v = '"' + "u".join(v) + '"'
                constructor += f"{k}={str(v)}, "
        return f'{self.__class__.__name__}({constructor.rstrip(", ")})'

    def __gt__(a, b):
        if isinstance(b, int):
            b = Duration(days=b)

        return a.just_days > b.just_days

    def __ge__(a, b):
        if isinstance(b, int):
            b = Duration(days=b)

        return a.just_days >= b.just_days

    def __lt__(a, b):
        if isinstance(b, int):
            b = Duration(days=b)

        return a.just_days < b.just_days

    def __le__(a, b):
        if isinstance(b, int):
            b = Duration(days=b)

        return a.just_days <= b.just_days

    @property
    def y(self):
        return self.years

    @property
    def Y(self):
        return self.years

    @property
    def year(self):
        return self.years

    @property
    def M(self):
        return self.months

    @property
    def m(self):
        return self.months

    @property
    def month(self):
        return self.months

    @property
    def D(self):
        return self.days

    @property
    def d(self):
        return self.days

    @property
    def day(self):
        return self.days


    def to_string(self):
        output = ''
        if self.years!=0:
            output+=f'{self.years:+}y'
        if self.months!=0:
            output+=f'{self.months:+}m'
        if self.days!=0:
            output+=f'{self.years:+}d'
        if self.bd:
            output+=f'{self.years:+}bd'
        if self.cals:
            output+=f'|{"u".join(self.cals)}'
        if self.modified:
            output+=f'/MOD'
        return output

DurationLike = (Duration,datetime.timedelta,dateutil.relativedelta.relativedelta)


if __name__ == "__main__":  # pragma: no cover
    from dateroll.ddh.ddh import ddh
    from dateroll import Date
    from dateroll import Duration

    # x = ddh('12/1/2023+12m/MOD')
    # x = ddh('6m+6m')
    # ddh('t+7dm')
    # ddh('t+7g')
    # x = ddh('6m+3m+1y+1/5/24')
    # print(x)
    # print(ddh('6m+7m+1y'))
    # x = ddh('2y-5m-4d')

    # print(Date(2024,5,5)-Date(2022,10,9))
    dur = ddh("4/15/24-1/15/24")
    print(dur.just_approx_days)

    ####### good test
    # x2 = Date(2022,10,9)
    # x = Date(2024,5,5)
    # x3 = x2-x
    # x4 = ddh('10/9/22 - 5/5/24')
    # print(x3)
    # print(x4)
    # print(x4==x3)
    #########

    # x = ddh('1/12/24+11m13d|NYuWE')
    # print(x)



"""
used to generate dateroll/sampledata/*.csv before packaging wheel
"""

import datetime
import os
import pathlib

from dateutil.relativedelta import relativedelta as rd
from workalendar.america.brazil import Brazil
from workalendar.asia.japan import JapanBank
from workalendar.europe.european_central_bank import EuropeanCentralBank
from workalendar.europe.united_kingdom import UnitedKingdom
from workalendar.usa import FederalReserveSystem

DEFAULT_YEAR_RANGE = 200


def generate_Workalendar(cls, n=DEFAULT_YEAR_RANGE):
    """
    use 3rd party library to get set of use federal holidays, return as set
    """
    y = datetime.datetime.today().year
    l = set()
    inst = cls()
    for year in range(y - n, y + n):
        hols = inst.holidays(year)
        l |= set(_[0] for _ in hols)
    return l


def generate_ALL_and_WE(n=DEFAULT_YEAR_RANGE):
    """
    generate set of all days in some "big range"
    generate set of holidays where holiday is saturday or sunday in some "big range"
    return both sets
    """
    t_start = datetime.date.today() - rd(years=n)
    t_end = datetime.date.today() + rd(years=n)

    t = t_start
    ALL = set()
    WE = set()
    while t < t_end:
        ALL |= {
            t,
        }
        if t.weekday() in (5, 6):
            WE |= {
                t,
            }
        t += rd(days=1)

    return ALL, WE


def generate_sample_data(cals, n=DEFAULT_YEAR_RANGE):
    ALL, WE = generate_ALL_and_WE(n)
    data = {
        "ALL": ALL,
        "WE": WE,
        "NY": generate_Workalendar(FederalReserveSystem, n=n),
        "EU": generate_Workalendar(EuropeanCentralBank, n=n),
        "LN": generate_Workalendar(UnitedKingdom, n=n),
        "BR": generate_Workalendar(Brazil, n=n),
    }

    for calendar_name, list_of_dates in data.items():
        filename = pathlib.Path(ROOT_DIR / "sampledata/" / f"{calendar_name}.csv")
        with open(filename, "w") as f:
            l = [i.isoformat() + "\n" for i in list_of_dates]
            f.writelines(l)


if __name__ == "__main__":
    ROOT_DIR = os.path.dirname(os.path.abspath(__file__))
    data = generate_sample_data('Na')
    for k, v in data.items():
        print(k, len(v))



import unittest

import code
import sys
from io import StringIO
import datetime
import dateutil.relativedelta
from dateroll import Date,Duration,ddh

class TestDuration(unittest.TestCase):
    @classmethod
    def setUpClass(cls): ...

    @classmethod
    def tearDownClass(self): ...


    def test_from_string(self):
        '''
        good string bad string
        '''
        x = Duration.from_string('+3m|NY')
        self.assertIsInstance(x,Duration)
        self.assertRaises(TypeError,lambda :Duration.from_string('garbage'))
        self.assertRaises(TypeError,lambda :Duration.from_string(3))
        
    def test_from_rd(self):
        '''
        good rd non-rd
        '''
        rd = dateutil.relativedelta.relativedelta(years=3,months=2,weeks=1,days=1)
        x = Duration.from_relativedelta(rd)
        self.assertIsInstance(x,Duration)
        self.assertRaises(TypeError,lambda :Duration.from_relativedelta('garbage'))

    def test_from_td(self):
        '''
        good td non-td
        '''
        td = datetime.timedelta(weeks=1,days=12)
        x = Duration.from_timedelta(td)
        y = Duration.from_relativedelta(td)
        self.assertIsInstance(x,Duration)
        self.assertIsInstance(y,Duration)
        self.assertRaises(TypeError,lambda :Duration.from_timedelta('garbage'))

    def test_period_units(self):
        '''
        properties
        '''

        y,m,w,d,bd = 1,2,3,4,5
        dur1 = Duration(y=y,m=m,w=w,d=d,bd=bd)
        dur2 = Duration()
      
        #year
        self.assertEqual(dur1.y,y)
        self.assertEqual(dur1.Y,y)
        self.assertEqual(dur1.year,y)
        self.assertEqual(dur1.years,y)

        #month
        self.assertEqual(dur1.m,m)
        self.assertEqual(dur1.M,m)
        self.assertEqual(dur1.month,m)
        self.assertEqual(dur1.months,m)

        #day
        self.assertEqual(dur1.d,d+7*w)
        self.assertEqual(dur1.D,d+7*w)
        self.assertEqual(dur1.day,d+7*w)
        self.assertEqual(dur1.days,d+7*w)
        
        #bd
        self.assertEqual(dur1.bd,bd)
        self.assertIsNone(dur2.bd)

    def test_roll_no_roll(self):
        '''
        test each perm, and bad versions, and neg
        '''
        dur = Duration(bd=1,modified=True)

        dur = Duration(bd=-3)
        

    def test_cals(self):
        cals1 = ['WE','NY']
        cals2 = 'WE'
        cals3 = 'WEuNY'
        good = ('NY','WE')

        cals4 = ['BAD','NY']
        cals5 = 'BAD'
        cals6 = 'NYuBAD'

        self.assertEqual(Duration(d=0,cals=cals1).cals,good)
        self.assertEqual(Duration(d=0,cals=cals2).cals,('WE',))
        self.assertEqual(Duration(d=0,cals=cals3).cals,good)

        self.assertRaises(ValueError,lambda : Duration(d=0,cals=cals4))
        self.assertRaises(ValueError,lambda : Duration(d=0,cals=cals5))
        self.assertRaises(ValueError,lambda : Duration(d=0,cals=cals6))
        

    def test_bdadj(self):
        '''
        test 0bd and >1bd and <1bd
        for each one of valid rolls
        '''

        # zero day adjustments

        # sun+0bd=mon
        # sun-0bd=fri
        # sun+1bd/F=mon
        # sat+1d/F=mon
        # error roll with 0bd that conflicts
        
        # non-zero bd follows
        # +1bd/F
        # +1bd/P
        # -1bd/F
        # -1bd/P

        # modified

        # +1bd/MF
        # +1bd/MP
        # -1bd/MF
        # -1bd/MP

    def test_delta(self):
        dur = Duration(
                    y=1,
                    m=1,
                    w=2,
                    d=11
        )
        rd = dur.relativedelta
        expected_rd = dateutil.relativedelta.relativedelta(years=1,months=1,weeks=2,days=11)
        self.assertEqual(rd,expected_rd)

    def test_adjust_bds(self):
        dur = Duration(bd=1)
        _,d = dur.adjust_bds(Date(2024,3,6))
        self.assertEqual(d,Date(2024,3,7))
    
    def test_adjust_bds(self):
        '''
            test business date adjustment which adjust date into business date
        '''
        d = Date(2024,3,6)
        expected_d = Date(2024,3,7)
        dur = Duration(bd=1)
        sign, newd = dur.adjust_bds(d)
        self.assertEqual(newd,expected_d)

        dur = Duration()
        sign, newd = dur.adjust_bds(d)
        self.assertEqual(newd,Date(2024,3,6))

        dur = Duration(modified=True)
        sign, newd = dur.adjust_bds(d)
        self.assertEqual(newd,Date(2024,3,6))

    def test__validated_periodunits(self):
        pass
    def test__validate_cals(self):
        '''
            test the validate cals where cals string should be 2,3 letters
        '''
        cals = ['NYYY','WE']
        with self.assertRaises(Exception) as cm:
            d = Duration(cals=cals)
        
        self.assertEqual(str(cm.exception),'Calendars must be 2 or 3 letter strings (not NYYY)')
        # validate cals must be str
        cals = [10]
        with self.assertRaises(Exception) as cm:
            d = Duration(cals=cals)
        self.assertEqual(str(cm.exception),'Calendars must be strings (not int)')
    def test__validate_adj_roll(self):
        pass

    def test_math(self):
        # coveraged by all dunder methods
        '''
            test the add/sub of dur and dur or dur and d
        '''
        dur2 = Duration(d=1)
        d1 = Date(2024,1,1)
        # newd1 = d1 + dur2
        
        # self.assertEqual(newd1,Date(2024,1,2))
        newd2 = d1 - dur2
        self.assertEqual(newd2,Date(2023,12,31))
        d1 = 10
        with self.assertRaises(NotImplementedError):
            d1 + dur2
        
        dur_cal = Duration(d=13,cals='WE')
        ndur = dur_cal + dur2
        self.assertEqual(ndur,Duration(years=0, months=0, days=14, modified=False, cals="WE"))
        dur_cal2 = Duration(d=1,cals='NY')
        ndur2 = dur_cal + dur_cal2
        self.assertEqual(ndur2,Duration(years=0, months=0, days=14, modified=False, cals="NYuWE"))

        bd1 = Duration(bd=1)
        bd2 = Duration(bd=2)
        bd3 = bd1 + bd2
        self.assertEqual(bd3,Duration(years=0, months=0, days=0, modified=False, bd=3, cals="WE"))
        rough1 = Duration(y=1,m=1,d=1)
        rough2 = Duration(y=-2,m=-2,d=-20,cals='WE')
        capt = StringIO()
        sys.stderr = capt
        r3 = rough1+rough2
        sys.stderr = sys.__stderr__
        
        self.assertEqual(r3,Duration(years=-1, months=-1, days=-19, modified=False, cals="WE"))

        mdur = Duration(d=13,modified=True)
        mdur2 = Duration(d=1,modified=True)
        self.assertEqual(mdur+mdur2,Duration(years=0, months=0, days=14, modified=True))
        
        nonedur1 = Duration(d=14)
        nonedur2 = Duration(d=1)
        self.assertEqual(nonedur1+nonedur2,Duration(years=0, months=0, days=15, modified=False))

        # test dur + rd, should not add rd, should return dur only
        rd3 = dateutil.relativedelta.relativedelta(days=3)
        rd7 = dateutil.relativedelta.relativedelta(days=7)
        dur4 = Duration(days=4)
        dur7 = Duration(days=7)
        dur7_from_math = dur4 + rd3
        self.assertEqual(dur7,dur7_from_math)
        self.assertEqual(dur7_from_math,rd7)

        # test dur +datelike and adjust business date if necessary
        dur = Duration(days=4)
        dt = Date(2024,1,1)
        x = dur+dt
        self.assertEqual(x,datetime.date(2024,1,5))




    def test___eq__(self):
        '''
            test equality
        '''
        from dateroll import ddh
        self.assertEqual(Duration(months=14),Duration(years=1,months=2))
        self.assertEqual(Duration(years=2),Duration(months=24))
        self.assertEqual(ddh('4/15/24-4/15/23'),Duration(years=1, months=0, days=0, modified=False, _anchor_start=Date(2023,4,15), _anchor_end=Date(2024,4,15), _anchor_months=12, _anchor_days=366, debug=False))
        self.assertEqual(ddh('4/15/24-4/12/23'),Duration(years=1, months=0, days=3, modified=False, _anchor_start=Date(2023,4,12), _anchor_end=Date(2024,4,15), _anchor_months=12, _anchor_days=369, debug=False))
        self.assertEqual(ddh('4/15/24-1/12/23'),Duration(years=1, months=3, days=3, modified=False, _anchor_start=Date(2023,1,12), _anchor_end=Date(2024,4,15), _anchor_months=15, _anchor_days=459, debug=False))
        self.assertEqual(ddh('6/15/24-1/15/24'),Duration(years=0, months=5, days=0, modified=False, _anchor_start=Date(2024,1,15), _anchor_end=Date(2024,6,15), _anchor_months=5, _anchor_days=152, debug=False))
        self.assertEqual(ddh('6/15/24-6/10/24'),Duration(years=0, months=0, days=5, modified=False, _anchor_start=Date(2024,6,10), _anchor_end=Date(2024,6,15), _anchor_months=0, _anchor_days=5, debug=False))
        self.assertEqual(ddh('4/15/24-1/12/23'),Duration(years=1, months=3, days=3, modified=False, _anchor_start=Date(2023,1,12), _anchor_end=Date(2024,4,15), _anchor_months=15, _anchor_days=459, debug=False))

        self.assertEqual(Duration(years=1).years,ddh('4/15/24-4/15/23').years)
        self.assertEqual(Duration(years=1,days=3).days,ddh('4/15/24-4/12/23').days)
        self.assertEqual(Duration(years=1,months=3,days=3).months,ddh('4/15/24-1/12/23').months)
        self.assertEqual(Duration(months=5).months,ddh('6/15/24-1/15/24').months)
        self.assertEqual(Duration(days=5).days,ddh('6/15/24-6/10/24').days)
        self.assertNotEqual(Duration(years=1,months=3,days=35).days,ddh('4/15/24-1/12/23').days)


        
        # pos testing
        self.assertEqual(Duration(days=5),Duration(days=5))
        self.assertEqual(Duration(days=5),dateutil.relativedelta.relativedelta(days=5))
        self.assertEqual(Duration(days=5),datetime.timedelta(days=5))

        # neg testing
        self.assertNotEqual(Duration(days=5),Duration(days=4))
        self.assertNotEqual(Duration(days=5),dateutil.relativedelta.relativedelta(days=4))
        self.assertNotEqual(Duration(days=5),datetime.timedelta(days=4))

        with self.assertRaises(TypeError):
            Duration(days=5)=='23423'

    def test___iadd__(self):
        dur = Duration(bd=5)
        dt = Date(2024,1,1)
        dur+=dt
        self.assertEqual(dur,datetime.date(2024,1,8))
    def test___init__(self):
        pass
    def test___isub__(self):
        dur = Duration(days=4)
        dt = Date(2024,1,1)
        dur-=dt
        self.assertEqual(dur,datetime.date(2023,12,28))
    def test___neg__(self):
        '''
            test __neg__ means -dur
        '''
        dur = Duration(days=4,modified=True)
        
        dur_neg = -dur
        self.assertEqual(dur.days,dur_neg.days * -1)
    def test___pos__(self):
        '''
            test __pos__ means +dur
        '''
        dur = Duration(days=4,modified=True)
        dur2 = +dur
        self.assertEqual(dur2,dur)
    
    def test_just_days(self):
        '''
            test the number of days between two anchor dates
        '''
        from dateroll import ddh
        dur = ddh('4/15/24-1/15/24')
        x = dur.just_days
        self.assertEqual(x,91)
        y = dur.just_approx_days
        self.assertEqual(y,91.3125)
        z = dur.just_exact_days
        self.assertEqual(z,91)


        z = Duration(months=3).just_days
        self.assertAlmostEqual(z,91.3125,2)

        z = Duration(year=1).just_days
        self.assertAlmostEqual(z,365.25,2)

        dur = Duration(bd=1)
        x = dur.just_days
        self.assertAlmostEqual(365.25/252,x,2)

        with self.assertRaises(ValueError):
            dur = Duration(years=1)
            dur.just_exact_days


    def test___radd__(self):
        pass
    def test___repr__(self):
        '''
            test repr of Duration instance
        '''
        dur = Duration(days=4,modified=True)
        dur2 = Duration(days=1,cals='NYuWE')
        repr(dur)
        repr(dur2)
    def test___rsub__(self):
        pass
    def test___sub__(self):
        pass

    def test_anchor(self):
        from dateroll import ddh
        dur = ddh('4/15/24-1/12/23')
        self.assertEqual(dur._anchor_months,15)
        self.assertEqual(dur._anchor_days,459)

    def test_bd0_and_roll(self):
        dur = Duration(bd=0,modified=True)
        self.assertEqual(dur.bd,0)
        dur = Duration(BD=0,modified=True)
        self.assertEqual(dur.bd,0)
        dur = Duration(BD=1,bd=2,modified=True)
        self.assertEqual(dur.bd,3)

    def test_mod(self):
        
        from dateroll import ddh

        # check which roll wins if adding left and right

        # same roll
        dur = ddh('+5bd/MOD')+ddh('+3bd/MOD')
        self.assertTrue(dur.modified)

        # left only
        dur = ddh('+5bd/MOD')+ddh('+3bd')
        self.assertTrue(dur.modified)

        # right only
        dur = ddh('+5bd')+ddh('+3bd/MOD')
        self.assertTrue(dur.modified)


        

        # both, 0 -> /F
        # batuuuu test again why it fails now!!! ###########
        dur = ddh('+5bd/MOD')+ddh('-5bd/MOD')
        self.assertTrue(dur.modified)

        # equal approx
        dur = ddh('+3m/MOD')+ddh('-3m/MOD')
        self.assertTrue(dur.modified)

        # minus with approx w/ modifier
        dur = ddh('+3m/MOD')+ddh('-5m/MOD')
        self.assertTrue(dur.modified)
        
        # minus with approx
        dur = ddh('+3m/MOD')+ddh('-5m/MOD')
        self.assertTrue(dur.modified)

    def test_compare(self):
        
        dur1 = Duration(months=5)
        dur2 = Duration(days=-3)

        # same
        self.assertEqual(dur1,dur1)
        self.assertNotEqual(dur1,dur2)
        self.assertEqual(dur2,-3)

        # >
        self.assertGreater(dur1,dur2)
        self.assertGreater(dur2,-5)

        # >=
        self.assertGreaterEqual(dur1,dur2)
        self.assertGreaterEqual(dur2,-3)
        
        # <
        self.assertLess(dur2,dur1)
        self.assertLess(dur2,-2)
        
        # <=
        self.assertLessEqual(dur2,dur1)
        self.assertLessEqual(dur2,-3)
    
    def test_staticMethods(self):
        '''
            passing durationlike would return duration
        '''
        x = Duration(bd=1)
        self.assertEqual(x,Duration.from_relativedelta(x))
        self.assertEqual(x,Duration.from_timedelta(x))
    
    def test_hash(self):
        x = Duration(bd=1)
        rs = hash(x)
        self.assertIsInstance(rs,int)

    def test_failMod(self):
        x = ddh('12/1/2023+12m/MOD')

    def test_sub(self):
        dur1 = Duration(days=12)
        dur2 = Duration(days=1)
        rs = dur2 - dur1
        
        self.assertEqual(rs,Duration(years=0, months=0, days=-11, modified=False, debug=False))




if __name__=='__main__':
    unittest.main()


import unittest

from tests.test_calendarmath import TestStringMathMethods
from tests.test_calendars import TestStringMethods
from tests.test_date import TestDate
from tests.test_ddh import TestDDH, TestsPracticalExamples
from tests.test_duration import TestDuration
from tests.test_operations import TestOperations
from tests.test_parser import TestParser
from tests.test_parsers import TestParsers
from tests.test_pretty import TestPretty
from tests.test_patterns import TestPatterns
from tests.test_schedule import TestSchedule
from tests.test_settings import TestSettings
from tests.test_usage import TestUsage, TestUsageMore

def test_suite():
    '''
        add test cases into suite
    '''
    suite = unittest.TestSuite()
    suite.addTest(unittest.makeSuite(TestStringMathMethods))
    suite.addTest(unittest.makeSuite(TestStringMethods))
    suite.addTest(unittest.makeSuite(TestDate))
    suite.addTest(unittest.makeSuite(TestDDH))
    suite.addTest(unittest.makeSuite(TestsPracticalExamples))
    suite.addTest(unittest.makeSuite(TestDuration))
    suite.addTest(unittest.makeSuite(TestOperations))
    suite.addTest(unittest.makeSuite(TestParser))
    suite.addTest(unittest.makeSuite(TestParsers))
    suite.addTest(unittest.makeSuite(TestPatterns))
    suite.addTest(unittest.makeSuite(TestPretty))
    suite.addTest(unittest.makeSuite(TestSchedule))
    suite.addTest(unittest.makeSuite(TestUsage))
    suite.addTest(unittest.makeSuite(TestUsageMore))
    suite.addTest(unittest.makeSuite(TestSettings))
    return suite

if __name__=='__main__':
    runner = unittest.TextTestRunner()
    suite = test_suite()
    runner.run(suite)


import warnings
import pathlib
import importlib.util
import code

path = pathlib.Path("~/.dateroll/settings.py").expanduser()

default_settings = {
    'debug':True,
    'convention':'MDY',
    'ie':'(]'
}

default_settings_validation ={
    'debug':(
        lambda x: isinstance(x,bool),
        TypeError('debug must be bool')
    ),
    'convention': (
        lambda x: isinstance(x,str) and x in ['YMD','MDY','DMY'],
        ValueError('must be one of "MDY", "DMY" or "YMD".')
    ),
    'ie': (
        lambda x: isinstance(x,str) and x in ['()','(]','[)','[]'],
        ValueError('must be one of (), (], [), or [] such that is the interval: (a,b])')
    )

}

class Settings:
    '''
    singleton whose attributes are stored in a plain .py file as name=value
    '''
    def __init__(self):
        '''
        if settings load and validate, if not load defaults and save
        '''
        self.load_default()

        if path.exists():
            self.load()
            self.validate()
        else:
            self.save()

    def load_default(self):
        '''
        update singleton with settings
        '''
        self.__dict__.update(default_settings)
    
    def load(self):
        '''
        try to load settings file, if corrupted replace with defaults
        '''
        try:
            spec = importlib.util.spec_from_file_location("module.name", path)
            mod = importlib.util.module_from_spec(spec)
            spec.loader.exec_module(mod)
            settings = mod.__dict__
            self.__dict__.update(settings)
        except Exception as e:
            msg = f'Settings corrupted in {path}, restoring defaults.'
            warnings.warn(msg)
            self.load_default()
            self.save()

    def save(self):
        '''
        save settings file
        '''
        path.parent.mkdir(parents=True, exist_ok=True)
        with path.open('w') as f:
            txt = f'# Dateroll settings file\n# {path}\n\n'
            for k,v in self.__dict__.items():
                if not k.startswith('__'):
                    txt += f'{k}={repr(v)}\n'
            f.write(txt)

    def validate(self):
        '''
        3 validate checks per setting / key is valid, type of value is valid, and value of value is valid
        '''
        # key check
        reset = False
        for k,v in self.__dict__.items():
            if k.startswith('__'):
                continue
            key_is_valid = k in default_settings
            type_is_valid = isinstance(v,type(default_settings.get(k,None)))
            missing_validation = (lambda x:False,ValueError(f'Setting {k} not found'))
            func_value_is_valid, exc = default_settings_validation.get(k,missing_validation)
            if  not (key_is_valid and type_is_valid and func_value_is_valid(v)):
                msg = f'Settings corrupted in {path}, restoring defaults.'
                warnings.warn(msg)
                self.load_default()
                break

    def __setattr__(self,k,v):
        '''
        add a new setting
        '''
        # check, set, save
        func_value_is_valid,exc = default_settings_validation[k]
        
        if not func_value_is_valid(v):
            raise exc
        
        super().__setattr__(k,v)
        self.save()

    def __repr__(self):
        ''' 
        show settings
        '''
        kwargs = ','.join([f'{k}={repr(v)}' for k,v in self.__dict__.items()])
        string = f'{self.__class__.__name__}({kwargs})'
        return string

                
settings = Settings()

if __name__=='__main__':  # pragma: no cover
    settings = Settings()


import os
import datetime
import pathlib
import shutil
import code
from dateroll.parser.parser import parse_to_dateroll, parse_to_native

"""
need daycounters

datroll.* better than native because you can add datestrings to date objects

ddh('t') +'3m' .... it's relaly close to the base but little extra.

"""

from dateroll.calendars.calendarmath import calmath
from dateroll.settings import settings
from dateroll.date.date import DateLike, Date
from dateroll.duration.duration import DurationLike, Duration

DEBUG = False

cals = calmath.cals


class ddh:
    calmath = calmath
    cals = cals
    def __new__(cls, o):
        if isinstance(o,str):
            obj = parse_to_dateroll(o)
        elif isinstance(o,DateLike):
            return Date.from_datetime(o)
        elif isinstance(o,DurationLike):
            obj = Duration.from_relativedelta(o)
        else:
            raise TypeError(f'ddh() cannot handle {type(o).__name__})')

        return obj

    @staticmethod
    def purge_all():
        """
        dangerous, deletes all calendars and lockfiles
        """
        p = pathlib.Path("~/.dateroll").expanduser()
        import glob
        files = glob.glob(str(p)+'/**/*',recursive=True)
        for file in files:
            if not file.endswith('lockfile'):
                if pathlib.Path(file).is_file():
                    os.remove(file)
        cals._purge_all()
        calmath._purge_all()
        



import unittest

class TestOperations(unittest.TestCase):
    @classmethod
    def setUpClass(cls): ...

    @classmethod
    def tearDownClass(self): ...

# no fol

# sunday
# monday
# monday as holiday
# tues
# wed
# wed as holiday
# thurs
# friday
# saturday

# x

# t-5bd
# t-4bd
# t-3bd
# t-2bd
# t-1bd
# t+0bd
# t+1bd
# t+2bd
# t+3bd
# t+4bd
# t+5bd

# fol


# american
# european
# intenrational

# YYMMDD
# MMDDYY
# DDMMYY
    
if __name__=='__main__':
    unittest.main()



# DATE_CONVERT_DICT = {
#     "1w": {"d": {"7d": "perfect"}},
#     "1m": {"d": {"30d": "approx"}, "w": {"4w": "approx"}},
#     "1q": {"d": {"90d": "approx"}, "w": {"12w": "approx"}, "m": {"3m": "perfect"}},
#     "1y": {"d": {"365d": "approx"}, "w": {"52w": "approx"}, "m": {"12m": "perfect"}, "q": {"4q": "perfect"}},
#     "7d": {"w": {"1w": "perfect"}},
#     "29d": {"m": {"1m": "approx"}},
#     "30d": {"m": {"1m": "approx"}},
#     "31d": {"m": {"1m": "approx"}},
#     "89d": {"q": {"1q": "approx"}},
#     "90d": {"q": {"1q": "approx"}},
#     "91d": {"q": {"1q": "approx"}},
#     "364d": {"y": {"1y": "approx"}},
#     "365d": {"y": {"1y": "approx"}},
#     "366d": {"y": {"1y": "approx"}},
#     "4w": {"m": {"1m": "approx"}},
#     "12w": {"q": {"1q": "approx"}},
#     "52w": {"y": {"1y": "approx"}},
#     "3m": {"q": {"1q": "perfect"}},
#     "12m": {"y": {"1y": "perfect"}},
#     "4q": {"y": {"1y": "perfect"}},
# }


# def get_numb_yr(per):
#     per = per.replace("+", "")
#     n_ptn = r"(\d+(?:\.\d+)?)"
#     yr_ptn = "[dwmqyDWMQY]"
#     numb = float(re.findall(n_ptn, per)[0])
#     yr = re.findall(yr_ptn, per)[0]
#     return numb, yr


# def period_add(converted_periods):
#     if len(converted_periods) == 1:
#         return converted_periods[0]
#     elif len(converted_periods) > 1:
#         _, per = get_numb_yr(converted_periods[0])  # since all values of list have same period
#         ls = [get_numb_yr(per)[0] for per in converted_periods]
#         numb = sum(ls)
#         return f"{numb}{per}"
#     else:
#         return converted_periods


# def date_convert_dict_helper(per, roughly):
#     converted_per = per
#     perfect_or_approx = "perfect"
#     numb, yr = get_numb_yr(per)
#     per2 = f"1{yr}"
#     if per2 in DATE_CONVERT_DICT.keys():
#         if roughly in DATE_CONVERT_DICT[per2].keys():
#             converted_per, perfect_or_approx = list(DATE_CONVERT_DICT[per2][roughly].items())[0]
#             numb2, yr2 = get_numb_yr(converted_per)
#             numb_last = numb * numb2
#             converted_per = f"{numb_last}{yr2}"
#     return converted_per, perfect_or_approx


# def date_convert_dict(per, roughly):
#     if per in DATE_CONVERT_DICT.keys():
#         if roughly in DATE_CONVERT_DICT[per].keys():
#             converted_per, perfect_or_approx = list(DATE_CONVERT_DICT[per][roughly].items())[0]
#         else:
#             converted_per, perfect_or_approx = date_convert_dict_helper(per, roughly)
#     else:
#         # numb,yr = get_numb_yr(per)

#         converted_per, perfect_or_approx = date_convert_dict_helper(per, roughly)
#     return converted_per, perfect_or_approx


# class Duration:  # look like a subclass of relativedelta
#     # dt_str some f(operatior,num,unit,cal_array,roll_function)
#     def __init__(self, date_period_string):
#         raise Exception
#         self.orig_date_period_string = date_period_string
#         # self.date_period_string,self.st,self.ed = dtYMD_convert(date_period_string)
#         self.date_period_string, self.days_, self.st, self.ed = self.per(date_period_string)
#         self.date_period_function = datePeriodStringToDatePeriod(self.date_period_string)

#     def convert(self, roughly=None):
#         date_period_strings = re.findall(r"\d+[dDwWmMqQsShHyY]", self.date_period_string)
#         converted_periods = []
#         for date_period_string in date_period_strings:
#             converted_per, perfect_or_approx = date_convert_dict(date_period_string, roughly)
#             converted_periods.append(converted_per)
#         converted_per = period_add(converted_periods)
#         if perfect_or_approx == "approx":
#             warnings.warn("roughly approximated, this may not be true as there is no anchor date")
#         return converted_per

#     @staticmethod
#     def from_relativedelta(rd_object):
#         '''
#         create Duration from dateutil.relativedelta.relativedelta
#         '''

#     @staticmethod
#     def from_timedelta(td_object):
#         '''
#         create Duration from datetime.timedelta
#         '''

#     @staticmethod
#     def from_string(td_object):
#         '''
#         create Duration from DurationString
#         '''


#     def cd(self, ie="[)"):
#         # self.stub='short',self.ret='l',self.monthEndRule='anniv',self.ie=NotImplementedError,self.dc=NotImplementedError
#         if self.st == None or self.ed == None:
#             raise Exception("both start date and end date should be given")

#         n = (self.ed - self.st).days
#         return n

#     def toDt(self):
#         if isinstance(self.rs, str):
#             self.rs = datetime.datetime.strptime(self.rs, "%Y%m%d")  # datetime.strptime(rs,'%Y%m%d')

#         # self.rs = self.rs.replace(hour=0, minute=0, second=0, microsecond=0)
#         from dateroll.date import Date
#         return Date(self.rs)

#     def today(self):
#         return self.date_period_function(datetime.date.today(), 1)

#     def __radd__(self, lhs):  # +
#         self.rs = self.date_period_function(lhs, 1)
#         return self.toDt()

#     def __rsub__(self, lhs):  # -
#         self.rs = self.date_period_function(lhs, -1)
#         return self.toDt()

#     def __add__(self, lhs):  # +
#         self.rs = self.date_period_function(lhs, 1)

#         return self.toDt()

#     def __sub__(self, lhs):  # -
#         self.rs = self.date_period_function(lhs, -1)
#         return self.toDt()

#     def __iadd__(self, lhs):  # +=
#         self.rs = self.date_period_function(lhs, 1)

#         return self.toDt()

#     def __isub__(self, lhs):  # -=
#         self.rs = self.date_period_function(lhs, -1)
#         return self.toDt()

#     def __str__(self):
#         return f"{self.orig_date_period_string}"

#     def __repr__(self):
#         return f'Duration("{self.orig_date_period_string}")'

#     def per(self, date_period_string):

#         try:
#             st, ed = re.findall(PTN, date_period_string)
#         except:
#             return date_period_string, "", None, None
#         # this part handles when date parse goes to format like 20220101-03,132022. Tried to other more robust methods, kinda stuck, later change the method
#         if "-" in st:
#             prs1 = st.split("-")

#             if "-" in ed:
#                 ed1 = "-".join([prs1[-1], ed])
#             elif "/" in ed:
#                 ed1 = "/".join([prs1[-1], ed])
#             elif "." in ed:
#                 ed1 = ".".join([prs1[-1], ed])
#             else:
#                 ed1 = "".join([prs1[-1], ed])
#             st = prs1[0]
#             ed = ed1
#         try:
#             st = parse(st)
#             ed = parse(ed)
#         except:
#             return date_period_string, "", None, None
#         m = 0
#         y = 0
#         d = []
#         dt = st
#         if st < ed:
#             dt_prev = dt
#             dt += relativedelta(months=1)

#             while dt <= ed:

#                 d.append((dt - dt_prev).days)
#                 m += 1
#                 if m == 12:
#                     m = 0
#                     y += 1
#                 dt_prev = dt
#                 dt += relativedelta(months=1)
#             if dt > ed:
#                 dt_prev = dt + relativedelta(months=-1)
#                 d.append((ed - dt_prev).days)
#             dur_string = f"-{y}y{m}m{d[-1]}d"
#             days_ = -sum(d)
#         elif st > ed:
#             dt_prev = dt
#             dt += relativedelta(months=-1)
#             while dt >= ed:

#                 d.append(abs((dt - dt_prev).days))
#                 m += 1
#                 if m == 12:
#                     m = 0
#                     y += 1
#                 dt_prev = dt
#                 dt += relativedelta(months=-1)
#             if dt < ed:
#                 d.append(abs((ed - dt_prev).days))
#             dur_string = f"{y}y{m}m{d[-1]}d"
#             days_ = sum(d)
#         else:
#             return "0y0m0d", 0, None, None
#         return dur_string, days_, st, ed

#     @property
#     def days(self):
#         n = Duration(self.date_period_string).convert("d")
#         n = float(n.replace("d", ""))
#         return n



import os
import datetime
import pathlib
import pickle
import time
import warnings
import code
import math


from dateroll.calendars.calendars import Calendars
from dateroll.utils import safe_open

PARENT_LOCATION = pathlib.Path.home() / ".dateroll/"
PARENT_LOCATION.mkdir(exist_ok=True)
MODULE_LOCATION = PARENT_LOCATION / "calendars/"
MODULE_LOCATION.mkdir(exist_ok=True)
DATA_LOCATION_FILE = MODULE_LOCATION / "compiled_cals"

WEEKEND_CALENDAR = 'WE'
DEFAULT_IE = "(]"


class CalendarMath:
    """
    CalendarMath is a singleton that holds "compiled" version of your Calendars
    Calendar is a dict-on-disk object that stores vectors of all holidays.
    CalendarMath stores calendar unions (cached), bd-counter dictionaries for your calendars, and more, and has a cached-on-disk backend as well.
    If a mutation of a calendar is detected, cached is purged and CalendarMath triggers a rebuild.

    CalendarMath operationsa are used by Duration and Date:

    add_bd(d,n,cals)    - add n bd's do to d on cals
    sub_bd(d,n,cals)    - sub n bd's to d on cals
    is_bd(d,cals)       - is d a bd on cals
    diff(d1,d2,cals,ie=IE_RULE) - how many bd's between d1 and d2 on cals and using inclusion/exclusion rules (] [) () or []
    next_bd(d1,cals)    - next business day (mod=True triggers modified following)
    prev_bd(d1,cals)    - previous business day (mod=True triggered modified previous)

    calenedar unions are handled automatically and cached, if a list of calendars is provided a new set of compiled calendars is created (doesn't impact Calendar's dict-on-disk representation

    """

    def __init__(self, home=DATA_LOCATION_FILE):
        self.home = home
        self.cals = Calendars()
        self.hash = self.cals.hash
        self.ALL = self.cals["ALL"]

        self.fwd = {}
        self.bck = {}

        self.cached_compile_all()

    def load_cache(self):
        '''
            self.home always exsist because load_cache called if it exists
        '''
        with open(self.home, "rb") as f:
            try:
                cached = pickle.load(f)
                return cached
            except:
                print(f'[dateroll] calendar math cache corrupted, cleaning.')
                os.remove(self.home)
        

    def save_cache(self):
        """
        cache internal state
        """
        cached = {
            "fwd": self.fwd,
            "bck": self.bck,
            "hash": self.hash,
        }
        with safe_open(self.home, "wb") as f:
            pickle.dump(cached, f)

    def cached_compile_all(self):
        """
        cached compile/load of calenders, uses temp file in user's home for cache.
        """
        if self.home.exists():
            # check cache on disk
            if not self.has_mutated:
                # load cache
                cached = self.load_cache()
                if cached is not None:
                    self.__dict__.update(cached)
                    return 
         
        # compile
        self.compile_all()

    def compile_all(self):
        print("[dateroll] compiling calendars")
        d = self.cals.copy()
        for k, v in d.items():
            if k == "WE":
                dates = v
                self.fwd[k], self.bck[k] = self.gen_dicts(k, v, self.ALL)

        self.save_cache()
        self.hash = self.cals.hash


    @property
    def has_mutated(self):
        '''
        only time "mutation" occurs is if an underlying calendar has changed
        all __setitem__ to a calendar will automatically trigger removal of the
        calendarmath cache file, thus it's existence confirms cache validity
        '''
        cache_is_valid = not self.home.exists()
        return cache_is_valid

    @property
    def recompile_if_mutated(self):
        if self.has_mutated:
            self.compile_all()

    @property
    def data_backend_present(self):
        """
        if any calendars are mutated the datemath backend is automatically removed by the Calendar class for saftey
        originally i checked "recompile_if_mutated" before calendar ops, but adds 89ms to do the unix checksome.
        This is a faster check for the existence of the data.
        """
        if not self.home.exists():
            self.recompile_if_mutated

    @staticmethod
    def gen_dicts(name, dates, all):
        """
        generates 4 dictionaries for a given calendar
        fwd[date]->#bds from inception
        bck[#bds]->date
        prev[date]->prev bus date
        next[date]->next bus date
        """
        # remove duplicates from dates
        dates = set(dates)
        cal = sorted(dates)

        fwd = {}
        bck = {}

        if len(cal) == 0:
            return fwd, bck

        last_cal = cal.pop(0)
        last_idx = 0
        # last_good = None
        for idx, dt in enumerate(sorted(all)):
            # goodbad = None

            if dt == last_cal:
                # gb = "holiday"
                fwd[dt] = last_idx
                # Batu: we should not add date here when it is holiday
                # if last_idx not in bck:
                #     bck[last_idx] = dt
                
                try:
                    last_cal = cal.pop(0)
                except:
                    break
            else:
                # is good bd
                last_idx += 1
                fwd[dt] = last_idx
                if last_idx not in bck:
                    bck[last_idx] = dt
                

        return fwd, bck

    def add_bd(self, d, n, cals):
        """
        add's n business days on cal to d
        careful, if n==0, B should auto backtrack, need to verify
        """
        
        # when not bd, we need to handle n; positive direction n=0-->1 and negative direction n=-1--->0
        # because of the property of fwd, bck dictionaries
        if not self.is_bd(d,cals):
            # make sure direction is positive
            if n==0 and math.copysign(1,n)==1:
                n = 1
            # when direction is negative
            elif n<0:
                n += 1
        cals = CalendarMath.reverse_calstring(cals)
        cal_name = self.union_key(cals)

        A = self.fwd[cal_name]
        B = self.bck[cal_name]
        if len(A)==0:
            raise ValueError('Please provide holidays')
        from dateroll.date.date import Date

        if isinstance(d, datetime.datetime):
            d = datetime.date(d.year, d.month, d.day)
        elif isinstance(d, Date):
            d = d.date
        elif isinstance(d, datetime.date):
            pass
        else:
            raise TypeError(f"Date must be date (got {type(d).__name__})")
            
        bd_index = A[d]
        new_bd_index = bd_index + n
        new_dt = B[new_bd_index]
        
        return new_dt

    def sub_bd(self, d, n, cals):
        """
        subtract (opposed of add)
        """
        if n<0:
            raise ValueError(f'n needs to be positive number')
        
        n = -1 * float(n)
        return self.add_bd(d, n, cals)

    def is_bd(self, d, cals):
        """am i business day?"""
        self.recompile_if_mutated

        """ cals can come in as str or list"""
        cals = CalendarMath.reverse_calstring(cals)

        for cal in cals:
            if cal not in self.cals:
                raise KeyError(f"There is no calendar {cal}")
            cal = self.cals[cal]
        
            if d in cal:
                
                return False
        
        return True

    def diff(self, t1, t2, cals, ie=DEFAULT_IE):
        """
        compute business days between two dates
        """
        cals = CalendarMath.reverse_calstring(cals)
        cal_name = self.union_key(cals)

        if t1==t2:
            return 0

        # exclude first or last based on ie (inclusion/exclusion rule)
        if ie[0]=='(':
            if t1 < t2:
                a = self.add_bd(t1,1,cals=cals)
            else:
                a = self.sub_bd(t1,1,cals=cals)
        else:
            a = t1

        if ie[1]==')':
            if t2 < a:
                b = self.add_bd(t2,1,cals=cals)
            else:
                b = self.sub_bd(t2,1,cals=cals)
        else:
            b = t2

        fwd = self.fwd[cal_name]
        n = fwd[b] - fwd[a] + 1

        return n

    def next_bd(self, d, cals):
        """
        the next business date from d on calendars cals
        """
        new_d = self.add_bd(d,1,cals)
        return new_d

    def prev_bd(self, d, cals):
        """
        the previous business date from d on calendars cals
        """
        new_d = self.sub_bd(d,1,cals)
        return new_d


    @staticmethod
    def reverse_calstring(cals):
        """
        cals can be:
            'WE'
            ['WE']
            'WEuNY'
            ['WE','NY']

        always return list of individuals (decompose unions)
        """
        if cals is None:
            cals = [WEEKEND_CALENDAR]  # define at top of file..future user might want WKD instad of WE

        if isinstance(cals, str):
            if "u" in cals:
                cals = tuple(sorted(cals.split("u")))
            else:
                cals = [cals]

        return cals

    def union_key(self, cals):
        """
        takes a cals as either  str of 1 cal or a list of str cals and return a sorted tuple, if empty list assume weekends
        
        """
        # cals always come as a tuple,list,set, no need to raise error here
        cal_union_key = 'u'.join(cals)

        if cal_union_key not in self.fwd:
            # union dates wasnt' cached, call _generate_union
            self._generate_union(cal_union_key)
        
        return cal_union_key
        

    def _generate_union(self, cal_union_key):
        """
        creates a union from sorted tuple of cals
        """
        unioned_dates = set()
        cals = CalendarMath.reverse_calstring(cal_union_key)
        for cal in cals:
            # saftey checks
            if cal not in self.cals:
                
                raise KeyError(f"There is no calendar {cal}")
            
            # the union operation
            unioned_dates |= set(self.cals[cal])

        # compile into large dict
        print(f"[dateroll] compiling new union [{cal_union_key}]")
        dict_tuple = self.gen_dicts(cal_union_key, unioned_dates, self.ALL)
        self.fwd[cal_union_key], self.bck[cal_union_key] = dict_tuple
        # save cache
        self.save_cache()

    def _purge_all(self):
        self.__init__()

    def __repr__(self):
        """
        Show names of cals and unions
        """
        return f'{self.__class__.__name__}(home="{self.home}")\nCals: {self.cals.keys()}\nUnions: {list(self.fwd.keys())}'


calmath = CalendarMath()

if __name__ == "__main__": # pragma: no cover

    import datetime
    import time

    from dateutil.relativedelta import relativedelta as rd

    t = datetime.date(2024, 2, 25)

    a = time.time()

    print(t)
    # calmath.add_bd(t,1,cals=['WE','NY'])

    t2 = calmath.add_bd(t, 1, cals=["WE"])
    print(t2)

    t2 = calmath.add_bd(t, 100, cals=["WE"])
    print(t2)
    print(calmath.diff(t, t2, cals=["WE"]))
    t2 = t + rd(days=31)
    print(calmath.diff(t, t2, cals=["WE"]))

    b = time.time()
    print((b - a) * 1000, "ms")

    


import datetime
import os
import pathlib
import tempfile
import unittest
import uuid
from unittest import expectedFailure
import dateroll.calendars.calendars as calendars
import code

from dateroll.settings import settings
from dateroll import ddh,cals
import dateroll
# import dateroll.duration.duration as durationModule


class TestDDH(unittest.TestCase):
    @classmethod
    def setUpClass(cls): ...

    @classmethod
    def tearDownClass(self): ...

    def testDateMathWithStr(self):
        """
        str->Date + str
        str->Date += str
        str + str->Date
        """

    def testDurationMathWithStr(self):
        """
        str->Duration + str
        str->Duration += str
        str + str->Duration

        """

    def testDateMathAdd(self):
        """
        str->Date + str->Date
        str->Date + str->Duration
        str->Duration + str->Date
        str->Duration + str->Duration
        """

    def testDateMathSub(self):
        """
        str->Date - str->Date
        str->Date - str->Duration
        str->Duration - str->Date
        str->Duration - str->Duration
        """

    def testSchedule(self):
        """
        str->Schedule (1bd,1d,1w,1m,1y)
        """

    def testPurge(self):
        '''
        purge all
        '''
        ddh.purge_all()
        base_cals = sorted(['FED', 'ECB', 'LN', 'WE', 'ALL', 'BR', 'NY'])
        self.assertEqual(sorted(cals.keys()),base_cals)

    def testConvention(self):
        original = settings.convention
        try: 
            # american
            settings.convention = 'MDY'
            _a = dateroll.Date(year=2024,month=12,day=1)
            a = ddh('12/1/24')
            self.assertEqual(_a,a)
            #european
            settings.convention = 'DMY'
            _b = dateroll.Date(year=2023,month=3,day=23)
            b = ddh('23/3/23')
            self.assertEqual(_b,b)
            #international
            settings.convention = 'YMD'
            _c = dateroll.Date(year=2022,month=10,day=5)
            c = ddh('20221005')      
            self.assertEqual(_c,c)
        finally:
            # back to original
            settings.convention = original

class TestsPracticalExamples(unittest.TestCase):
    def testNothing(self):
        """
        empty string
        """
        with self.assertRaises(dateroll.parser.parsers.ParserStringsError):
            ddh("")

    def testDate(self):
        """
        test for date strings
        """
        assert ddh("t") == datetime.date.today()
        assert ddh("5/5/05") == datetime.date(2005, 5, 5)

    def testDuration(self):
        """
        test for duration strings
        """

    def testSchedule1(self):
        """
        test for schedule strings
        """
        assert ddh("t") == datetime.date.today()

    def test_durationLike(self):
        '''
            pass durationlike into ddh
        '''
        x = ddh(datetime.timedelta(days=10))
        self.assertEqual(x,dateroll.Duration(years=0, months=0, days=10, modified=False, debug=False))
    
    def test_badObj(self):
        '''
            pass bad instace to rayse TypeError
        '''
        
        with self.assertRaises(TypeError):
            ddh(10)
if __name__ == "__main__":
    unittest.main()

    # x = "2/24/24+3bd|WEuNY"
    # print(x, "ddh=", ddh(x))
    # # print("answer:",ddh('3/1/24'))
    # # print("answer:",ddh('t'))
    # # print("answer:",ddh('-3m|NY'))
    # # print("answer:",ddh('9m|NY+ -11m'))
    # # print("answer:",ddh('9m|NY + 11m0bd|EU'))
    # # print("answer:",ddh('-6m7y2d5q8w|NYuEUuWEuBRuMXuARuCONZ/MF'))
    # # print("answer:",ddh('9m+6m7y2d5q8w|NYuEUuWEuBRuMXuARuCONZ/MF'))
    # # print("answer:",ddh('t+1bd'))
    # # print("answer:",ddh('t-5bd'))
    # # print("answer:",ddh('t-0bd'))
    # # print("answer:",ddh('t+3m'))



import warnings
import code
import unittest
import os

import dateroll.parser.parsers as parsers
import datetime
from dateroll.settings import Settings, path
from dateroll.date.date import Date
from dateroll.duration.duration import Duration


class TestSettings(unittest.TestCase):
    @classmethod
    def setUpClass(cls): ...

    @classmethod
    def tearDownClass(self): ...

    def test_init(self):
        '''
            test init of settings
        '''
        settings = Settings()
        settings.load_default()
        
        self.assertEqual(settings.convention,'MDY')

        os.remove(path)
        settings = Settings()
        self.assertEqual(settings.convention,'MDY')
    
    def test_load(self):
        settings = Settings()
        settings.load_default()
        settings.load()
        self.assertEqual(settings.convention,'MDY')
    
    def test_validate(self):
        settings = Settings()
        settings.convention = 'YMD'
        settings.validate()
        self.assertEqual(settings.convention,'YMD')
        settings.convention = 'MDY'
        settings.validate()

    def test_setattr(self):
        '''
        set bad attributes
        '''
        settings = Settings()

        with self.assertRaises(Exception):
            settings.debug = 'hello'

        with self.assertRaises(Exception):
            settings.convention = 'ABC'
    
    def test_repr(self):
        settings = Settings()
        x = repr(settings)

    def test_bad_settings(self):
        ...
        # bad key
        with self.assertWarns(Warning):
            with open(path,'w') as f:
                f.write('bad_setting=123')
            settings = Settings()

        # syntax error
        with self.assertWarns(Warning):
            with open(path,'w') as f:
                f.write('/')
            settings = Settings() 


if __name__=='__main__':
    unittest.main()


import code
import unittest

import dateroll.parser.parsers as parsers
import datetime
from dateroll.settings import settings
from dateroll.date.date import Date
from dateroll.duration.duration import Duration


class TestParsers(unittest.TestCase):
    @classmethod
    def setUpClass(cls): ...

    @classmethod
    def tearDownClass(self): ...

    def test_parseTodayString(self):
        '''
            get today date, we pass t,today,t0 and get the date in a string format
        '''
        #store convention
        orig = settings.convention
        try:

            #american
            settings.convention = 'MDY'
            expected_mdy = datetime.date.today().strftime('%m/%d/%Y')
            t = parsers.parseTodayString('t')
            self.assertEqual(t.replace('-','/'),expected_mdy)

            #european
            settings.convention = 'DMY'
            expected_dmy = datetime.date.today().strftime('%d/%m/%Y')
            t = parsers.parseTodayString('t')
            self.assertEqual(t.replace('-','/'),expected_dmy)

            #international
            settings.convention = 'YMD'
            expected_ymd = datetime.date.today().strftime('%Y/%m/%d')
            t = parsers.parseTodayString('t')
            self.assertEqual(t.replace('-','/'),expected_ymd)

            #negative testing
            self.assertTrue('cannot do with a more stricter parser than dateutil.parser.parse')

        finally:
            #reset convention
            settings.convention = orig
    
    def test_parseDateString(self):
        '''
            test the parse date string based on convention
        '''
        #store convention
        orig = settings.convention
        letters = [chr(i) for i in range(65, 65 + 26)]
        def gen(): yield letters.pop(0)
        try:

            #american
            settings.convention = 'MDY'
            a = '03/08/2024'
            l,s = parsers.parseDateString(a,gen)
            
            b = list(l.values())[0].to_string().replace('-','/')
            self.assertEqual(a,b)

            #european
            settings.convention = 'DMY'
            a = '08/03/2024'
            l,s = parsers.parseDateString(a,gen)
            b = list(l.values())[0].to_string().replace('-','/')
            self.assertEqual(a,b)

            #international
            settings.convention = 'YMD'
            a = '2024/03/08'
            l,s = parsers.parseDateString(a,gen)
            b = list(l.values())[0].to_string().replace('-','/')
            self.assertEqual(a,b)

            #negative testing
            self.assertTrue('cannot do with a more stricter parser than dateutil.parser.parse')

        finally:
            #reset convention
            settings.convention = orig
    
    def test_process_duration_match(self):
        '''
            test duration match
        '''
        s = ('','+','1','y','0','q','0','m','1','w','1','d','0','bd','WE','NY','BR','ECB','FED','LN','','','MOD')
        dur = parsers.process_duration_match(s)
        self.assertEqual(dur,Duration(years=1, months=0, days=8, modified=True, cals="BRuECBuFEDuLNuNYuWE"))

        s2 = ('','-','1','y','0','q','0','m','1','w','1','d','0','bd','WE','NY','BR','ECB','FED','LN','','','MOD')
        dur2 = parsers.process_duration_match(s2)
        self.assertEqual(dur2,Duration(years=-1, months=0, days=8, modified=True, cals="BRuECBuFEDuLNuNYuWE"))

        s3 = ('','*','1','y','0','q','0','m','1','w','1','d','0','bd','WE','NY','BR','ECB','FED','LN','','','P')
        with self.assertRaises(Exception) as cm:
            parsers.process_duration_match(s3)
        self.assertEqual(str(cm.exception),'Unknown operator')
        
        # no roll but it will have roll P
        s4 = ('','-','1','y','0','q','0','m','1','w','1','d','0','bd','WE','NY','BR','ECB','FED','LN','','','')
        dur4 = parsers.process_duration_match(s4)
        self.assertEqual(dur4,Duration(years=-1, months=0, days=8, modified=False, cals="BRuECBuFEDuLNuNYuWE"))
    def test_parseCalendarUnionString(self):
        '''
            test cal union string e.g, NYuBR --> [NY,BR]
        '''
        
        s = 'NYuWE'
        s2 = parsers.parseCalendarUnionString(s)
        self.assertEqual(s2,('NY','WE'))

        # wrong string raise error
        with self.assertRaises(Exception) as cm:
            parsers.parseCalendarUnionString('ab')
        self.assertEqual(str(cm.exception),'ab not a valid calendar string')

    def test_parseDurationString(self):
        '''
            parse duration string
        '''
        letters = [chr(i) for i in range(65, 65 + 26)]
        def gen(): yield letters.pop(0)
        s = '1y3m4d'
        s2 = parsers.parseDurationString(s,gen)
        
        self.assertEqual(s2,({'A': Duration(years=1, months=3, days=4, modified=False, debug=False)}, '+A'))
    
    def test_parseDateMathString(self):
        '''
            test parse date math string, +X-X
        '''

        # good
        self.assertEqual(parsers.parseDateMathString("A",{'A':4}),4)
        self.assertEqual(parsers.parseDateMathString("-A",{'A':4}),-4)
        self.assertEqual(parsers.parseDateMathString("+A",{'A':4}),4)
        self.assertEqual(parsers.parseDateMathString('A+B', {'A':4,'B':5}),9)
        self.assertEqual(parsers.parseDateMathString("A+B-C",{'A':4,'B':3,'C':1}),6)
        self.assertEqual(parsers.parseDateMathString("A+B-C-D",{'A':4,'B':3,'C':1,'D':1}),5)

        # bad
        with self.assertRaises(Exception) as cm:
            parsers.parseDateMathString("A+B", {'A':4})
    
        # bad
        with self.assertRaises(Exception) as cm:
            parsers.parseDateMathString('B',{'A':4,'B':5})        
    
    def test_parsersConvention(self):
        s = settings
        
        







if __name__=='__main__':
    unittest.main()


import fcntl
import pathlib

XPRINT_ON = False

DEBUG_COLORS = {
    "blue": "\033[94m",
    "cyan": "\033[96m",
    "green": "\033[92m",
    "yellow": "\033[33m",
    "red": "\033[31m",
    "end": "\033[0m",
    "gray": "\x1b[38;5;239m",
}


def color(s, color):
    return DEBUG_COLORS[color] + str(s) + DEBUG_COLORS["end"]


def xprint(*args, **kwargs):  # pragma:no cover
    if XPRINT_ON:
        _color = kwargs.get("color", "yellow")
        if "lbl" in kwargs:
            lbl = kwargs.get("lbl")
            b = kwargs.get("before", "x")
            a = kwargs.get("after", "x")
            b = str(b)
            a = str(a)
            a_star = ""
            for _a, _b in zip(a, b):
                if _a == _b:
                    a_star += color(_a, "green")
                else:
                    a_star += color(_a, "blue")
            b = color(b, "green")
            s = f"{color('before',_color)}: {b}, {color('after',_color)}: {a_star}"
            print(color(f"[debug] {lbl:>12} ", _color), s)
        else:
            args = [color(a, _color) for a in args]
            print(color("[debug]", _color), *args, **kwargs)


def combine_none(a, b):
    if a is None and b is None:
        return None
    a = [] if a is None else a
    b = [] if b is None else b
    return tuple(sorted(set(a) | set(b)))


def add_none(a, b, dir=1):
    if a is None and b is None:
        return None
    else:
        a = 0 if a is None else a
        b = 0 if b is None else b
        return a + b * dir


class safe_open:
    """
    use separate write lockfile to lock/unlock (fcntl removes dealock if pid w/ lock dies)
    lock is removed if pid dies
    """

    def __init__(self, path, mode="r"):
        """
        open lockfile and attempt to lock, will block
        """
        self.path = pathlib.Path(path)
        self.mode = mode
        self.pathlock = self.path.with_suffix(".lockfile")
        self.lockfile = open(self.pathlock, "w")
        fcntl.lockf(self.lockfile, fcntl.LOCK_EX)

    def __enter__(self):
        """
        open user file and send it to them
        """
        self.file = open(self.path, self.mode)
        return self.file

    def __exit__(self, exc_type, exc_val, exc_tb):
        """
        release lock, then close both lockfile and user file
        do not delete lockfile (even if no exc)
        """
        fcntl.lockf(self.lockfile, fcntl.LOCK_UN)
        self.lockfile.close()
        self.file.close()


convention_map = {"YMD": r"%Y-%m-%d", "DMY": r"%d-%m-%Y", "MDY": r"%m-%d-%Y"}



import datetime
import os
import pathlib
import tempfile
import unittest
import uuid
import code
from unittest import expectedFailure

from dateroll.calendars.calendarmath import CalendarMath,DATA_LOCATION_FILE
from dateroll.date.date import Date
import dateroll.calendars.calendars as calendars



class TestStringMathMethods(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.filename_base = (
            pathlib.Path(tempfile.gettempdir())
            / f"dateroll.testing.self.cals.{uuid.uuid4()}"
        )
        cls.cals = CalendarMath(home=cls.filename_base)

    @classmethod
    def tearDownClass(self):
        filename = str(self.filename_base)
        os.remove(filename)

    def test_addbd(self):
        """
        test add business days

        without mod:
            non-hol+0bd|WE
            hol+0bd|WE
            non-hol+0bd|WEuNY
            hol+0bd|WEuNY
            hol+1000bd
            nonhol+1000bd

        measure perf, put a threshold 

        """
        # without mod
        nonhol = Date(2023,1,3)
        x1 = self.cals.add_bd(nonhol,0,'WE')
        self.assertEqual(x1,Date(2023,1,3))
        hol = Date(2024,1,1)
        x2 = self.cals.add_bd(hol,0,['WE'])
        self.assertEqual(x2,Date(2024,1,1))
        x3 = self.cals.add_bd(nonhol,0,['WE','NY'])
        self.assertEqual(x3,Date(2023,1,3))

        x4 = self.cals.add_bd(hol,10000,'WE')

        self.assertEqual(x4,Date(2062,5,1))
        x5 = self.cals.add_bd(nonhol,10000,'WE')
        self.assertEqual(x5,Date(2061,5,3))
        

    def test_subbd(self):
        """
        just 1 or 2 from add, uses same mechanics

        """
        # without mod
        nonhol = Date(2023,1,3)
        hol = Date(2024,1,1)

        x1 = self.cals.sub_bd(nonhol,0,'WE')
        self.assertEqual(x1,Date(2023,1,3))
        
        x2 = self.cals.sub_bd(hol,0,['WE'])
        self.assertEqual(x2,Date(2024,1,1))
        x3 = self.cals.sub_bd(nonhol,0,['WE','NY'])
        self.assertEqual(x3,Date(2023,1,3))
        
        x4 = self.cals.sub_bd(hol,0,['WE','NY'])
        
        self.assertEqual(x4,Date(2023,12,29))
        x5 = self.cals.sub_bd(hol,10000,'WE')
        x6 = self.cals.sub_bd(nonhol,10000,'WE')
        self.assertEqual(x5,Date(1985,9,2))
        self.assertEqual(x6,Date(1984,9,4))



    def test_isbd(self):
        """
        test is business day

        test a holiday is bd on non-union
        test a holiday in a is bd on union of a u b
        test a holiday in b is bd on union of a u b u c
        test nonholiday is not bd on non-union
        test nonholiday is not bd on union

        measure perf, put a threshold
        """
        hol = Date(2024,1,1)
        is_bd = self.cals.is_bd(hol,'WE')
        self.assertTrue(is_bd is True)
        is_bd = self.cals.is_bd(hol,['NY','WE'])
        self.assertFalse(is_bd is True)

        nonhol = Date(2024,1,4)
        is_bd = self.cals.is_bd(nonhol,'WE')
        self.assertTrue(is_bd is True)
        is_bd = self.cals.is_bd(nonhol,'WEuNY')
        self.assertTrue(is_bd is True)
        is_bd = self.cals.is_bd(nonhol,'WEuNYuBR')
        self.assertTrue(is_bd is True)


    def test_diffbd(self):
        """
        change global setting:  settings.ie='(])' do the test, then try again
                
        one liner test code:
        
        python -c "from dateroll import *; settings.ie='[]'; b=ddh('3/22/24'); a= ddh('3/18/24'); d=b-a; d.cals=['NY']; print(d.just_bds)"

        pick a calendar where on WEuNY you have 5 dates:

            a. MTWRF No hol
            b. M holiday, TWU no hol
            c. F holiday, TWU no hol
            d. MF holiday, TWU no hol
            e. W holiday, MTU no hol

        for each of the 5 above, you have 4 cases to test, so 20 test cases:

        MTWRF No hol
            () - 3
            (] - 4
            [) - 4
            [] - 5
        M holiday, TWU no hol
            () - 3
            (] - 4
            [) - 3
            [] - 4
        F holiday, TWU no hol
            () - 3
            (] - 3
            [) - 4
            [] - 4
        MF holiday, TWU no hol
            () - 3
            (] - 3
            [) - 3
            [] - 3
        W holiday, MTU no hol
            () - 2
            (] - 3
            [) - 3
            [] - 4

        e.g.

        """
    
        cals = 'NY'

        a = Date(2024,1,1)
        b = Date(2024,1,12)
        x = Date(2024,2,29)
        y = Date(2024,3,5)
        # [b,a) is here
        b_minus_a = self.cals.diff(a,b,cals)
        self.assertEqual(b_minus_a,9)
        a_minus_b = self.cals.diff(b,a,cals)
        self.assertEqual(a_minus_b,-9)

        #[y,x) here
        x_minus_y = self.cals.diff(x,y,cals)
        self.assertEqual(x_minus_y,3)
        x_minus_a = self.cals.diff(a,x,cals)
        self.assertEqual(x_minus_a,41)
        x_minus_b = self.cals.diff(x,b,cals)
        self.assertEqual(x_minus_b,-32)
        y_minus_a = self.cals.diff(a,y,cals)
        self.assertEqual(y_minus_a,44)
        y_minus_b = self.cals.diff(b,y,cals)
        self.assertEqual(y_minus_b,35)
        raise Exception('reimpliement test bd with i rules working using real test cases')

    

    def test_dataBackendPresent(self):
        '''
            test data_backend_present function
        '''
        p = pathlib.Path('NonExist')
        if p.exists():
            os.remove(p)
        calmath = self.cals
        hash = calmath.hash
        calmath.home = pathlib.Path('NonExist')
        calmath.data_backend_present
        hash2 = calmath.hash
        self.assertEqual(hash,hash2)

    
    def test_emptyCalDates(self):
        '''
            when the dates are empty, e.g cals['AA'] = []
        '''
        cal = self.cals.cals
        if 'AA' in cal.keys():
            del cal['AA']
        cal['AA'] = []
        calmath = self.cals
        nonhol = Date(2023,1,3)
        with self.assertRaises(Exception) as cm:
            self.cals.add_bd(nonhol,2,'AA')
        self.assertEqual(str(cm.exception),'Please provide holidays')

    
    def test_datetime(self):
        '''
            when pass datetime and date
        '''
        calmath = self.cals
        nonhol = datetime.datetime(2023,1,3)
        d = self.cals.add_bd(nonhol,1,None)
        self.assertEqual(d,Date(2023,1,4))

        # date
        nonhol2 = datetime.date(2023,1,3)
        d = self.cals.add_bd(nonhol2,1,'WE')
        self.assertEqual(d,Date(2023,1,4))
    
    def test_NonDate(self):
        '''
            when add non date should raise typeerror
        '''
        calmath = self.cals
        dt = 10
        with self.assertRaises(Exception) as cm:
            self.cals.add_bd(dt,2,'WE')
        self.assertEqual(str(cm.exception),'Date must be date (got int)')
    
    def test_calNotInCalKeys(self):
        '''
            when calendar is not in the Calendar instance
        
        '''

        nonhol = datetime.datetime(2023,1,3)
        
        with self.assertRaises(Exception) as cm:
            self.cals.add_bd(nonhol,1,'AB')
        self.assertEqual(str(cm.exception),"'There is no calendar AB'")
    
    def test_NonStringCal(self):
        '''
            calendar must be a string
        '''

        nonhol = datetime.datetime(2023,1,3)
        with self.assertRaises(Exception) as cm:
            self.cals.add_bd(nonhol,2,[10])
        
    def test_negativeN(self):
        '''
            when we try to pass n<0 in sub_bd, should raise error
        '''
        calmath = self.cals
        nonhol = datetime.datetime(2023,1,3)
        with self.assertRaises(Exception) as cm:
                    self.cals.sub_bd(nonhol,-1,'WE')
                
        self.assertEqual(str(cm.exception),'n needs to be positive number')
    def test_repr(self):
        calmath = self.cals
        x = repr(calmath)

    def test_nextBd(self):
        '''
            find the next bd
        '''

        nonhol = datetime.datetime(2023,12,29)
        hol = Date(2023,12,31)
        d = self.cals.next_bd(hol,'NYuWE')
        d2 = self.cals.next_bd(nonhol,'NYuWE')
        self.assertEqual(d,Date(2024,1,2))
        self.assertEqual(d2,Date(2024,1,2))
    
    def test_prevBd(self):
        '''
        
            find the previous bd
        '''
        
        
        nonhol = datetime.datetime(2024,1,2)
        hol = Date(2024,1,1)
        d = self.cals.prev_bd(hol,'NYuWE')
        d2 = self.cals.prev_bd(nonhol,'NYuWE')
        self.assertEqual(d,Date(2023,12,29))
        self.assertEqual(d2,Date(2023,12,29))
    
    def test_compileAll(self):
        self.cals.cached_compile_all()
    
    def test_generate_union(self):
        '''
            when calendar does not exist, raise keyError
        '''
        
        with self.assertRaises(KeyError):
            self.cals._generate_union('BAD')



    



if __name__ == "__main__":
    unittest.main()



import re
import unittest
import io
from contextlib import redirect_stdout


from dateroll import *

class TestPretty(unittest.TestCase):

    def testPretty(self):
        
        # 1 cal
        with io.StringIO() as buf, redirect_stdout(buf):
            s = ddh('5/5/95+5d').cal
            output = buf.getvalue()
            self.assertGreater(len(output),300) # range is for variable ansi chars
            self.assertLess(len(output),400)
       
        # 2 cal
        with io.StringIO() as buf, redirect_stdout(buf):
            s = ddh('5/5/95+10y').cal
            output = buf.getvalue()
            self.assertGreater(len(output),600)  # range is for variable ansi chars
            self.assertLess(len(output),1000)

        # 2 cal
        with io.StringIO() as buf, redirect_stdout(buf):
            s = ddh('3/4/24-10y').cal
            output = buf.getvalue()
            self.assertGreater(len(output),600)  # range is for variable ansi chars
            self.assertLess(len(output),1000)

        # 1 date 1 cal
        with io.StringIO() as buf, redirect_stdout(buf):
            s = ddh('5/5/95').cal
            self.assertGreater(len(output),300) # range is for variable ansi chars
            self.assertLess(len(output),1025)

        # 1 schedule 2 cal
        with io.StringIO() as buf, redirect_stdout(buf):
            s = ddh('t,t+5y,1m').cal
            self.assertGreater(len(output),600)  # range is for variable ansi chars
            self.assertLess(len(output),1000)

if __name__ == "__main__":
    unittest.main()



import datetime
import unittest
import code

from dateroll import utils
from dateutil.relativedelta import relativedelta
from dateroll.settings import settings
from dateroll import Date, Duration


class TestDate(unittest.TestCase):
    @classmethod
    def setUpClass(cls): ...

    @classmethod
    def tearDownClass(self): ...

    def test___cmp__(self):
        """
        test compares
        """

        # dateroll.Date with dateroll.Date
        a = Date(2024, 12, 5)
        b1 = Date(2024, 12, 5)
        b2 = Date(2024, 12, 6)
        self.assertEqual(a, b1)
        self.assertNotEqual(a, b2)

        # dateroll.Date with datetime.date
        b1 = datetime.date(2024, 12, 5)
        b2 = datetime.date(2024, 12, 6)
        self.assertEqual(a, b1)
        self.assertNotEqual(a, b2)

        # dateroll.Date with datetime.datetime
        b1 = datetime.datetime(2024, 12, 5)
        b2 = datetime.datetime(2024, 12, 6)
        self.assertEqual(a, b1)
        self.assertNotEqual(a, b2)

    def test_to_conversions(self):
        """
        dateroll.Date to datetime.datetime and datetime.date
        """

        a = Date(1900, 1, 1)
        b = datetime.date(1900, 1, 1)
        c = datetime.datetime(1900, 1, 1)

        self.assertEqual(b, a.date)
        self.assertEqual(c, a.datetime)

    def test_from_dateti(self):
        """conversion from"""

        ref = Date(1900, 1, 1)
        a = datetime.date(1900, 1, 1)
        b = datetime.datetime(1900, 1, 1)

        d1 = Date.from_datetime(a)
        self.assertEqual(ref, d1)

        d2 = Date.from_datetime(b)
        self.assertEqual(ref, d2)

        d3 = Date.from_string("1/1/1900")
        self.assertEqual(ref, d3)

    def test_is_bd(self):
        """
        check if day is a business day given a specific calendar
        """
        sunday = Date.from_string("3/3/24")
        monday = Date.from_string("3/4/24")
        christmas = Date.from_string("12/25/23")

        self.assertFalse(sunday.is_bd(cals="WEuLN"))
        self.assertTrue(monday.is_bd(cals="WEuNYuBR"))
        self.assertFalse(christmas.is_bd(cals="WEuNYuBR"))

    def test_conversions_out(self):
        """
        various properties
        """
        d = Date.from_string("3/3/24")
        # iso
        self.assertEqual(d.iso, "2024-03-03")

        # xls
        self.assertEqual(d.xls, 45354)

        # unix
        ts = d.datetime.timestamp()
        self.assertEqual(d.unix, ts)

        # #dotw
        dotw = d.dotw
        self.assertEqual(dotw, "Sun")

        # woty
        woty = d.woty
        self.assertEqual(woty, 9)

    def test_operations(self):
        """
        add, iadd, sub, rsub
        """
        d1 = Date(2024, 1, 3)
        d2 = Date(2024, 4, 3)
        d3 = datetime.date(2024,4,3)
        d4 = datetime.datetime(2024,4,3)
        dur = Duration(m=3)
        rd = relativedelta(months=3)
        td = datetime.timedelta(days=91)
        str_d1 = "1/3/24"
        str_dur = "+3m"
        int_dur = 91

        # add
        self.assertRaises(TypeError, lambda: d1 + d2)
        self.assertRaises(TypeError, lambda: d1 + str_d1)
        self.assertEqual(d1 + str_dur, d2)
        self.assertEqual(d1 + dur, d2)
        self.assertEqual(d1 + int_dur, d2)
        self.assertEqual(d1 + rd, d2)
        self.assertEqual(d1 + td, d2)
        self.assertEqual(d3-d2,Duration())

        self.assertRaises(TypeError,lambda: d1+3.0)
        self.assertRaises(TypeError,lambda: 3.0-d1)

        # # sub
        dur91d = Duration(y=0,m=3,w=0,d=0,_anchor_start=Date(2024,1,3),_anchor_end=Date(2024,4,3),)
        
        self.assertEqual((d2 - d1), dur91d)
        self.assertEqual(d2 - dur, d1)
        self.assertEqual(d2 - str_d1, dur91d)
        self.assertEqual(d2 - str_dur, d1)
        self.assertEqual(d2 - int_dur, d1)
        self.assertEqual(d2 - rd, d1)
        self.assertEqual(d2 - td, d1)
        self.assertEqual(d2 - d3,Duration(days=0))
        self.assertEqual(d2-d2,d3-d2)
        self.assertEqual(d2-d2,d4-d2)
        self.assertEqual(d2-d2,d2-d4)

        self.assertRaises(TypeError,lambda: dur-d1)
        self.assertRaises(TypeError,lambda: 3.0-d1)
        self.assertRaises(TypeError,lambda: d1-3.0)
        self.assertRaises(TypeError,lambda: d2-3.0)

        # iadd
        _d1 = d1
        try:
            _d1 += d2
            assert False
        except TypeError:
            assert True

        _d1 = d1
        _d1 += dur
        self.assertEqual(_d1, d2)

        _d1 = d1
        try:
            _d1 += str_d1
            assert False
        except TypeError:
            assert True

        _d1 = d1
        _d1 += int_dur

        self.assertEqual(_d1, d2)

        _d1 = d1
        _d1 += rd
        self.assertEqual(_d1, d2)

        _d1 = d1
        _d1 += td
        self.assertEqual(_d1, d2)

        # isub
        _d2 = d2
        _d2 -= d1
        self.assertEqual(_d2, d1)

        _d2 = d2
        _d2 -= str_d1
        self.assertEqual(_d2, d1)

        _d2 = d2
        _d2 -= dur
        self.assertEqual(_d2, d1)

        _d2 = d2
        _d2 -= str_dur
        self.assertEqual(_d2, d1)

        _d2 = d2
        _d2 -= int_dur
        self.assertEqual(_d2, d1)

        _d2 = d2
        _d2 -= td
        self.assertEqual(_d2, d1)

        _d2 = d2
        _d2 -= rd
        self.assertEqual(_d2, d1)

    def test_repr(self):
        '''
            test the repr of date
        '''
        a = Date(2024, 12, 5)
        rs = repr(a)
        fmt = utils.convention_map[settings.convention]
        dstr = a.strftime(fmt)
        self.assertEqual(rs,f'Date(year=2024,month=12,day=5)')

    def test_from_string(self):
        '''
            test from string, pass Date or bad instance to raise TypeError
        '''

        x = Date(2023,1,1)
        self.assertEqual(Date.from_string(x),x)
        
        with self.assertRaises(TypeError):
            Date.from_string(10)
    
    def test_fromDatetime(self):
        '''
            pass bad instance will raise Type error
        '''
        with self.assertRaises(TypeError):
            Date.from_datetime(10)
    
    def test_today(self):
        self.assertEqual(Date.today(),datetime.date.today())
    
    def test_src(self):
        '''
            test src which prints calendar
        '''
        ... # coveraged in test_pretty


        

if __name__ == "__main__":
    unittest.main()



import datetime
import fcntl
import glob
import hashlib
import os
import pathlib
import pickle
import time
import code
from random import choice

from dateroll.utils import safe_open

ROOT_DIR = pathlib.Path(__file__).parents[2]
PARENT_LOCATION = pathlib.Path.home() / ".dateroll/"
PARENT_LOCATION.mkdir(exist_ok=True)
MODULE_LOCATION = PARENT_LOCATION / "calendars/"
MODULE_LOCATION.mkdir(exist_ok=True)
DATA_LOCATION_FILE = MODULE_LOCATION / "holiday_lists"
SAMPLE_DATA_PATH = ROOT_DIR / "dateroll" / "sampledata" / "*.csv"

SLEEP_TIMES = [25 / 1000, 50 / 1000, 100 / 1000, 300 / 1000]


INCEPTION = datetime.date(1824,2,29)


def load_sample_data():
    files = glob.glob(str(SAMPLE_DATA_PATH))
    data = {}
    for file in files:
        name = pathlib.Path(file).stem
        with open(file) as f:
            ls = f.readlines()
            ld = {}
            for i in ls:
                dt = datetime.date(int(i[0:4]), int(i[5:7]), int(i[8:10]))
                if dt > INCEPTION:
                    ld[dt]=True
            data[name] = ld
    return data


class Drawer:
    def __init__(self, cals):
        
        self.path = pathlib.Path(cals.home)
        self.cals = cals

    def __enter__(self):
        if self.cals.hash == self.cals.db_hash:
            return self.cals.db

        try:
            with safe_open(self.path, "rb") as f:
                self.data = pickle.load(f)
                self.cals.db_hash = self.cals.hash
                self.cals.db = self.data
                return self.cals.db
        except:
            print(f"[dateroll] no saved calendars, loading sample data")
            data = load_sample_data()
            self.cals.db = data
            with safe_open(self.path, "wb") as f:
                pickle.dump(self.cals.db, f)

            return self.cals.db

    def __exit__(self, exc_type, exc_val, exc_tb):
        if exc_type is None:
            if self.cals.write:
                with safe_open(self.path, "wb") as f:
                    pickle.dump(self.cals.db, f)
                self.cals.write = False
        else:

            return True


class Calendars(dict):
    """
    dict-like dictionary of calendars
    """

    def __init__(self, home=DATA_LOCATION_FILE):
        self.home = str(home)  # on disk location
        self.db_hash = None  # initial hash
        self.db = {}  # cache
        self.write = False  # sentinel to invalidate cache

        with Drawer(self) as db:
            pass


    def keys(self):
        '''
            return date keys
        '''
        with Drawer(self) as db:
            return list(db.keys())

    @property
    def hash(self):
        """
        generate hash
        """
        result = -1
        filenames = glob.glob(f"{self.home}")
        if len(filenames) == 1:
            filename = filenames[0]
            with open(filename, "rb") as f:
                result = hashlib.md5(
                    f.read(),
                ).hexdigest()

        return result

    def __setitem__(self, k, v):
        from dateroll.calendars.calendarmath import \
            DATA_LOCATION_FILE as calendar_math_file

        if calendar_math_file.exists():
            os.remove(calendar_math_file)

        # key must be 2-3 letter string in uppercase
        if not isinstance(k, str):
            raise Exception(f"Cal name must be string (got {type(k).__name__})")
        if len(k) < 2 or len(k) > 3:
            raise Exception(f"Cal name be 2 or 3 charts (got {len(k)})")
        if not k.isupper():
            raise Exception(f"Cal name must be all uppercase")
        # value must be a list of dates
        if not isinstance(v, (set, list, tuple)):
            raise Exception(
                f"Cal values must be a set/list/tuple (got {type(v).__name__})"
            )

        processed = []
        for i in v:
            
            if isinstance(i, datetime.datetime):
                dt = datetime.date(i.year, i.month, i.day)
            elif isinstance(i, datetime.date):
                dt = i
            else:
                raise Exception(
                    f"All cal dates must be of dateroll.Date or datetime.date{{time}} (got {type(i).__name__})"
                )
            if dt >= INCEPTION:
                
                processed.append(dt) 

        self.write = True
        if k in self.db.keys():
            raise Exception(
                f"{k} exists already, delete first.if you want to replace."
            )
        with Drawer(self) as db:
            s = {k:True for k in list(sorted(list(set(processed))))}
            db[k] = s

    def __getitem__(self, k):
        return self.get(k)

    def __getattr__(self, k):
        """
        allows for dot notation
        """
        if k in ("hash", "home"):
            result = super().__getattribute__(k)
        else:
            result = self.get(k)
        return result

    def __contains__(self, k):
        with Drawer(self) as db:
            return str(k) in db

    def __delitem__(self, k):
        self.write = True
        with Drawer(self) as db:
            del db[k]

    def get(self, k):
        with Drawer(self) as db:
            result = db[k]
        return result

    def clear(self):
        self.write = True
        with Drawer(self) as db:
            db.clear()

    def __repr__(self):
        self.info
        return f'{self.__class__.__name__}(home="{self.home}")'

    def copy(self):
        with Drawer(self) as db:
            return db.copy()

    def _purge_all(self):
        self.__init__()

    @property
    def info(self):
        pattern = lambda a, b, c, d: f"{a:6}|{b:>8}|{c:12}|{d:12}"
        with Drawer(self) as db:
            stats = []
            print(pattern("name", "#dates", "min date", "max date"))
            print(pattern("-" * 6, "-" * 8, "-" * 12, "-" * 12))
            for i in db.keys():
                
                l = db.get(i)
                if len(l) > 0:
                    n = len(l)
                    mn = min(l)
                    mx = max(l)
                else:
                    n, mn, mx = 0, None, None
                print(pattern(str(i), str(n), str(mn), str(mx)))


if __name__ == '__main__': # pragma: no cover
    ...
    



import datetime
import calendar

from dateroll.settings import settings

calendar.setfirstweekday(calendar.SUNDAY)

from dateroll.utils import color


def replacer(i, c, txt1, cals):
    bef = str(i).rjust(3)
    aft = color(bef, c)
    txt2 = txt1.replace(bef, aft)
    return txt2


def pretty_between_two_dates(dt1, dt2, cals, calmath):
    """
    experimental feature

    print a calendar or calenders for set of debugging:
        if start and end date fall in same month show calendar for month
        if standard and end date fall in two different months show calendars for starting month and ending month

        color skipped days as cyan
        color start/end days as blue
        color holidays/weekends as grey
    """
    if dt2 > dt1:
        _1, _2 = dt1, dt2
    else:
        _1, _2 = dt2, dt1

    print('from',_1,'to',_2, 'cals',cals)
    
    y1,m1,d1 = _1.year,_1.month,_1.day
    y2,m2,d2 = _2.year,_2.month,_2.day

    c1 = calendar.month(y1, m1)
    c1s = str(c1).splitlines()
    c1s[-1] = c1s[-1].ljust(len(c1s[-2]))
    cal1 = "\n".join([" " + i[:] for i in c1s][1:])

    if m1 == m2 and y1 == y2:
        # 1 month cal mode
        for i in range(0, 32):
            try:
                _ = datetime.date(year=y1, month=m1, day=i)
                ishol = not calmath.is_bd(_, cals)
            except:
                ishol = False

            if i == d1 or i == d2:
                cal1 = replacer(i, "blue", cal1, cals)
            elif ishol:
                cal1 = replacer(i, "gray", cal1, cals)
            elif i > d1 and i < d2:
                cal1 = replacer(i, "cyan", cal1, cals)

        res = color(f"        {_1.strftime('%b')} {y1}", "yellow")
        for idx, (i, j) in enumerate(zip(cal1.splitlines(), cal1.splitlines())):
            if idx > 0:
                res += "\n" + f"  {i} "

    else:
        # 2 month cal mode
        for i in range(0, 32):
            try:
                _ = datetime.date(year=y1, month=m1, day=i)
                ishol = not calmath.is_bd(_, cals)
            except:
                ishol = False
            if i == d1:
                cal1 = replacer(i, "blue", cal1, cals)
            elif i >= d1 and not ishol:
                cal1 = replacer(i, "cyan", cal1, cals)
            elif ishol:
                cal1 = replacer(i, "gray", cal1, cals)

        c2 = calendar.month(y2, m2)
        cal2 = "\n".join([" " + i for i in str(c2).splitlines()][1:])
        for i in range(0, 32):
            try:
                _ = datetime.date(year=y1, month=m1, day=i)
                ishol = not calmath.is_bd(_, cals)
            except:
                ishol = False
            if i == d2:
                cal2 = replacer(i, "blue", cal2, cals)
            elif i <= d2 and not ishol:
                cal2 = replacer(i, "cyan", cal2, cals)
            elif ishol:
                cal2 = replacer(i, "gray", cal2, cals)

        res = color(
            f"        {_1.strftime('%b')} {y1}                 {_2.strftime('%b')} {y2}                ",
            "yellow",
        )
        for idx, (i, j) in enumerate(zip(cal1.splitlines(), cal2.splitlines())):
            res += "\n" + f"  {i}  {' '+j}"
            prev = len(i)

    return res



import datetime
import dateutil
import dateutil.relativedelta
import dateutil.rrule
import code

# from dateroll.date.date import Date
import dateroll.date.date as dateModule
# from dateroll.duration.duration import Duration
import dateroll.duration.duration as durationModule
from dateroll.parser import parsers
from dateroll.schedule.schedule import Schedule


class ParserError(Exception): ...


class Parser:
    """

    Parser input/output:

        dateroll types:

        str -> dateroll.Date
        str -> dateroll.Duration
        str -> dateutil.Schedule

        or to python native types (with flag):

        str -> datetime.date
        str -> datetime.timedelta
        str -> dateutil.relativedelta.relativedelta
        str -> dateutil.rrule.rrule


    str "string" must one of the following formats:

        1 - TodayString (today's date)
        2 - DateString (represents a date)
        3 - DurationString (represents a duration)
        4 - DateMathString (e.g. date + duration, date - date, etc...)
        5 - DateScheduleString (represents a trip of end-to-end dates)

        ^^ details on each is in the subsequent parsing sub-fucntion

    Patterns:

        TodayString

            a   TodayString                         ->   DateString

        DateString:

            b   DateString                          ->   Date | datetime.date | datetime.datetime

        DurationString:

            c   DurationString                      ->   Duration | datetime.timedelta | datetime.relativedelta.relativedelta

        DateMathString (addition subtypes:)

            d   DateString      +   Duration        ->   Date | datetime.date | datetime.datetime
            e   DurationString  +   DateString      ->   Date | datetime.date | datetime.datetime
            f   DurationString  +   DurationString  ->   Duration | datetime.timedelta | datetime.relativedelta.relativedelta
                Duration        +   DateString      ->   raise Exception
                DateString      +   DateString      ->   raise Exception

        DateMathString (subtraction subtypes:)

            g   DateString      -   DateString      ->   Duration | datetime.timedelta | datetime.relativedelta.relativedelta
            h   DateString      -   Duration        ->   Date | datetime.date | datetime.datetime
            i   DurationString  -   DurationString  ->   Duration | datetime.timedelta | datetime.relativedelta.relativedelta
                Duration        -   DateString      ->   raise Exception
                DurationString  -   DateString      ->   raise Exception

        DateScheduleString:

            j    DateString      ,   DateString      ,    Schedule | dateutil.rrule.rrule
            k    DateString      ,   DateMathString  ,    Schedule | dateutil.rrule.rrule
            l    DateMathString  ,   DateString      ,    Schedule | dateutil.rrule.rrule
                 DateMathString  ,   DateMathString  ,    Schedule | dateutil.rrule.rrule


    """

    native_date = datetime.date
    native_delta = dateutil.relativedelta.relativedelta

    def __new__(
        self, string, use_native_types=False,
    ):
        """
        Algorithm works left to right implicitly:
        Check for how many parts (1 or 3 allowed, "," separator)
        For each part
            1 - convert TodayStrings into DateStrings
            2 Match DateString's -> Date
            3 Match DurationString's -> Duration
            4 Match DateMathString -> Date or Duration
        """

        if not isinstance(string, str):
            raise ParserError("Must be string")

        if use_native_types: 
            raise NotImplementedError("only dateroll types for now")

        self.use_native_types = use_native_types
        
        
        part = Parser.parse_maybe_many_parts(string)
        return part

    @classmethod
    def parse_one_part(cls, untouched):
        letters = [chr(i) for i in range(65, 65 + 26)]
        def gen():yield letters.pop(0)
        # 1
        notoday = parsers.parseTodayString(untouched)

        # 2
        dates, nodates = parsers.parseDateString(notoday,gen)
        # print('s before/after:', untouched,nodates)

        # 3
        durations, nodatesordurations = parsers.parseDurationString(nodates,gen)
        # print('s before/after:', nodates,nodatesordurations)
        dates_durations = {**dates,**durations}
        
        # 4
        # print('before pdms',nodatesordurations)
        processed_answer = parsers.parseDateMathString(
            nodatesordurations, dates_durations
        )
        return processed_answer

    @classmethod
    def parse_maybe_many_parts(cls, s):
        parts = s.split(",")
        if not s:
            TypeError("No empty strings")

        num_parts = len(parts)

        if num_parts == 1:
            part = parts[0]
            date_or_period = cls.parse_one_part(part)
            return date_or_period

        elif num_parts == 3:
            _maybe_start, _maybe_stop, _maybe_step = parts

            maybe_start = cls.parse_one_part(_maybe_start)
            maybe_stop = cls.parse_one_part(_maybe_stop)
            maybe_step = cls.parse_one_part(_maybe_step)

            if isinstance(maybe_start, dateModule.Date):
                start = maybe_start
            else:
                raise TypeError("Start of generation must be a valid Date")

            if isinstance(maybe_stop, dateModule.Date):
                stop = maybe_stop
            else:
                raise TypeError("Stop of generation must be a valid Date")

            if isinstance(maybe_step, durationModule.Duration):
                step = maybe_step
            else:
                raise TypeError("Step of generation must be a valid Duration")

            sch = Schedule(start=start, stop=stop, step=step, origin_string=s)
            return sch

        else:
            # Must
            raise Exception(
                f"String must contain either 1 or 3 parts (not {num_parts})"
            )

        


def parse_to_native(string):
    return Parser(string, use_native_types=True)


def parse_to_dateroll(string):
    return Parser(string)



import unittest
import code

from dateroll.settings import settings
from dateroll.parser.parser import Parser, parse_to_native
from dateroll.date.date import Date
from dateroll.duration.duration import Duration
class TestParser(unittest.TestCase):
    @classmethod
    def setUpClass(cls): ...

    @classmethod
    def tearDownClass(self): ...

    def test_new(self):
        '''
            tests things in __new__
        '''

        # only accepts string
        with self.assertRaises(Exception) as cm:
            pars = Parser(10)
        self.assertEqual(str(cm.exception),'Must be string')

        
        with self.assertRaises(NotImplementedError):
            Parser('s',use_native_types=True)
    
    def test_parse_one_part(self):
        '''
            test parse one part
        '''
        original = settings.convention
        try:
            settings.convention = 'YMD'
            
            dt = Parser.parse_one_part('20240101')
            self.assertEqual(dt,Date(2024,1,1))

            dur = Parser.parse_one_part('3m')
            
            self.assertEqual(dur,Duration(years=0, months=3, days=0, modified=False, debug=False))

            dt_dur = Parser.parse_one_part('20240101+3m')
            
            self.assertEqual(dt_dur,Date(2024,4,1))
        finally:
            settings.convention = original
    
    def test_maybe_many_parts(self):
        '''
            test parse maybe many parts accepts 1 part or 3 parts (start,stop,step)
        '''


        # 1 part
        dt = Parser.parse_maybe_many_parts('20240101')
        self.assertEqual(dt,Date(2024,1,1))

        # 3 parts - valid schedule generation
        dts = Parser.parse_maybe_many_parts('20240101,20240201,1bd')
        self.assertEqual(dts.dates[0],Date(2024,1,1))
        self.assertEqual(dts.dates[-1],Date(2024,2,1))
        self.assertEqual(len(dts),24)

        # 3 parts 1st part wrong
        with self.assertRaises(TypeError):
            dt = Parser.parse_maybe_many_parts('3m,t,1m')

        # 3 parts 2nd part wrong
        with self.assertRaises(TypeError):
            dt = Parser.parse_maybe_many_parts('t,3m,1m')

        # 3 parts 3rd part wrong
        with self.assertRaises(TypeError):
            dt = Parser.parse_maybe_many_parts('t,t+3m,t-1d')
            
        # parts stub
        dt = Parser.parse_maybe_many_parts('t,t+2m15d,1m')

        # 2 parts are wrong
        with self.assertRaises(Exception):
            dt = Parser.parse_maybe_many_parts('t,t+3m')
        

        # use_native NotImplmented for now
        with self.assertRaises(NotImplementedError):
            dt = parse_to_native('t')

    

if __name__=='__main__':
    unittest.main()


import datetime
import os
import pathlib
import tempfile
import unittest
import uuid
import code
import sys
from io import StringIO
from unittest import expectedFailure

from dateroll.calendars.calendars import Calendars,Drawer
import dateroll


class TestStringMethods(unittest.TestCase):
    @classmethod
    def setUpClass(cls):
        cls.filename_base = (
            pathlib.Path(tempfile.gettempdir())
            / f"dateroll.testing.{uuid.uuid4()}.cals"
        )
        cls.cals = Calendars(home=cls.filename_base)

    @classmethod
    def tearDownClass(self):
        filename = str(self.filename_base)
        os.remove(filename)

    def test_keys(self):
        """
        ensure the keys method is returning a list-like object
        """
        k = self.cals.keys()
        assert isinstance(k, list)

    @expectedFailure
    def test_tests1(self):
        self.cals["aaaa"] = []

    @expectedFailure
    def test_tests2(self):
        self.cals["aa"] = []

    @expectedFailure
    def test_tests3(self):
        self.self.cals["AA"] = []

    def test_tests4(self):
        del self.cals['AA']
        self.cals["AA"] = []

    def test_tests5(self):
        del self.cals["AA"]

    def test_tests6(self):
        self.cals["AA"] = ()

    @expectedFailure
    def test_tests7(self):
        self.cals["AA"] = (1,)

    @expectedFailure
    def test_tests8(self):
        self.cals["AA"] = (1,)

    def test_tests9(self):
        dates = [datetime.date.today()]
        self.cals["BBB"] = dates
        assert list(self.cals["BBB"].keys()) == dates
    
    def test_exitReturnTrue(self):
        '''
            context manager fails in the __exit__
        '''
        cals = Calendars()
        with Drawer(cals) as db:
            try:
                db.update('Hello')
            except Exception as e:
                exc_type,exc_value,exc_traceback = sys.exc_info()
        self.assertEqual(exc_type.__name__,'ValueError')
        
    def test_contains(self):
        '''
            test contains has a k in its dict keys
        '''
        cals = Calendars()
        cond = 'FED' in cals
        self.assertTrue(cond)
    
    def test_setitemkeyNonString(self):
        '''
            when setitem key is Non String should raise error
        '''
        with self.assertRaises(Exception) as context:
            self.cals[30] = []
        self.assertEqual(str(context.exception),'Cal name must be string (got int)')
    
    def test_setitemValWrong(self):
        '''
            when setitem val is not list,tuple,set
        '''
        with self.assertRaises(Exception) as context:
            self.cals['AA'] = 10
        
        self.assertEqual(str(context.exception),'Cal values must be a set/list/tuple (got int)')
    
    def test_seitemValDate(self):
        '''
            when we assign the list of datetime.date
        '''
        self.cals['AA'] = [dateroll.Date(2023,1,1)]
        self.assertTrue(dateroll.Date(2023,1,1) in self.cals['AA'])
    
    def test_seitemValDateTime(self):
        '''
            when we assign the list of datetime.datetime
        '''
        del self.cals['AA']
        self.cals['AA'] = [datetime.datetime(2023,1,1)]
        self.assertTrue(dateroll.Date(2023,1,1) in self.cals['AA'])

    def test_seitemValDateClass(self):
        '''
            when we assign the list of Date Class
        '''
        x = dateroll.Date(2020,1,1)
        self.cals['AB'] = [x]
        
        self.assertTrue(dateroll.Date(2020,1,1) in self.cals['AB'])
    
    def test_setitemValWrong2(self):
        '''
            when we assign the val which is not date related
        '''
        del self.cals['AA']
        with self.assertRaises(Exception) as context:
            self.cals['AA'] = ['20230101']
        self.assertEqual(str(context.exception),'All cal dates must be of dateroll.Date or datetime.date{time} (got str)')
    
    def test_setitemKeyExistsAlready(self):
        '''
            when the key already exists, it wont set the new value
        '''
        cals = Calendars()
        with self.assertRaises(Exception) as cm:
            cals['FED'] = [dateroll.Date(2024,2,1)]
        self.assertEqual(str(cm.exception),'FED exists already, delete first.if you want to replace.')
    
    def test_getattrNotHashHome(self):
        '''
            when the key of getattr not in (hash,home)
        '''
        cals = Calendars()
        x = cals.FED
        self.assertIsInstance(x,dict)
        self.assertTrue(len(x)>0)
    
    def test_getattrHashHome(self):
        '''
            when the key of getattr in (hash,home)
        '''
        cals = Calendars()
        x = cals.__getattr__('home')
        # dont want to put home/batu in the result
        x = x.split('dateroll')[-1]
        self.assertEqual(x,'/calendars/holiday_lists')

    def test_delitem(self):
        '''
            test delitem of calendar
        '''
        self.cals['AA'] = [dateroll.Date(2023,2,1)]
        self.assertTrue('AA' in self.cals)
        del self.cals['AA']
        self.assertFalse('AA' in self.cals)
    
    def test_repr(self):
        '''
            test __repr__ of calendar
        '''
        expected_str = 'Calendars(home="/tmp/dateroll.'
        self.assertEqual(repr(self.cals).split('testing')[0],expected_str)
    
    def test_copy(self):
        '''
            the copy of the db should be as same as cals
        '''
        cals = Calendars()
        copy_db = cals.copy()
        self.assertTrue(copy_db,self.cals.db)
    
    def test_info(self):
        '''
            test the info of calendar
        '''
        cals = Calendars()
        capt = StringIO()
        sys.stdout = capt

        cals.info
        sys.stdout = sys.__stdout__

        printed_output = capt.getvalue()
        expected_output = 'name  |  #dates|min date    |max date    \n------|--------|------------|------------\nFED   |    4556|1824-01-01  |2223-12-25  \nECB   |    2400|1824-01-01  |2223-12-26  \nLN    |    3951|1824-01-01  |2223-12-26  \nWE    |   41742|1824-02-29  |2224-02-28  \nALL   |  146097|1824-02-29  |2224-02-28  \nBR    |    3586|1824-01-01  |2223-12-25  \nNY    |    4556|1824-01-01  |2223-12-25  \n'
        self.assertTrue(printed_output,expected_output)
    
    def test_clear(self):
        '''
            clear the db
        '''
        self.assertTrue(len(self.cals.db)>0)
        self.cals.clear()
        # dict becomes 0
        self.assertTrue(len(self.cals.db)==0)
    
    def test_emptyDBInfo(self):
        '''
            clear the db and show the info
        '''
       
        capt = StringIO()
        sys.stdout = capt

        self.cals.info
        sys.stdout = sys.__stdout__

        printed_output = capt.getvalue()
        expected_output = 'name  |  #dates|min date    |max date    \n------|--------|------------|------------\nFED   |    4556|1824-01-01  |2223-12-25  \nECB   |    2400|1824-01-01  |2223-12-26  \nLN    |    3951|1824-01-01  |2223-12-26  \nWE    |   41742|1824-02-29  |2224-02-28  \nALL   |  146097|1824-02-29  |2224-02-28  \nBR    |    3586|1824-01-01  |2223-12-25  \nNY    |    4556|1824-01-01  |2223-12-25  \n'
        self.assertTrue(printed_output,expected_output)
        self.cals.clear()
        self.cals['AC'] = []
        capt = StringIO()
        sys.stdout = capt

        self.cals.info
        sys.stdout = sys.__stdout__
        printed_output_empty = capt.getvalue()
        expected_output_empty = 'name  |  #dates|min date    |max date    \n------|--------|------------|------------\nAC    |       0|None        |None        \n'
        self.assertTrue(printed_output_empty,expected_output_empty)

    


        


        

if __name__ == "__main__":
    unittest.main()



import datetime
import os
import pathlib
import tempfile
import unittest
import uuid
from unittest import expectedFailure
import pandas as pd
import code
import time

from dateroll.utils import color

import dateroll.calendars.calendars as calendars
from dateroll import ddh,cals
import dateroll.date.date as dateModule
from dateroll.settings import settings
import dateroll.duration.duration as durationModule
from tests.test_data.test_data import next_d,prev_d

def handle_sample_data_dates(x,inc,sign='+'):
    t = x['today']
    modified = x['modified']
    if modified:
        s = f'{t}{sign}{inc}|NYuWE/MOD'
    else:
        s = f'{t}{sign}{inc}|NYuWE'
    
    orig = settings.convention
    settings.convention = 'YMD'
    res = ddh(s)
    settings.convention = orig
    return res

class TestUsage(unittest.TestCase):
    @classmethod
    def setUpClass(cls): ...

    @classmethod
    def tearDownClass(self): ...

    def test_next_bd01(self):
        '''
            add 0 bd and 1bd from today
        
        '''
        df = pd.DataFrame(next_d)
        # print('in ddh')
        bd0 = df.apply(lambda x:handle_sample_data_dates(x,'0bd','+'),axis=1)
        bd0 = [l.strftime('%Y-%m-%d') for l in bd0.tolist()]
        expected0 = df['0bd'].tolist()
        expected1 = df['1bd'].tolist()
        self.assertEqual(bd0,expected0)
        bd1 = df.apply(lambda x:handle_sample_data_dates(x,'1bd','+'),axis=1)
        bd1 = [l.strftime('%Y-%m-%d') for l in bd1.tolist()]
        self.assertEqual(bd1,expected1)
        
    
    def test_prev_bd01(self):
        '''
            sub 0bd and 1bd from today
        '''
        df = pd.DataFrame(prev_d)
        bd0 = df.apply(lambda x:handle_sample_data_dates(x,'0bd','-'),axis=1)
        bd0 = [l.strftime('%Y-%m-%d') for l in bd0.tolist()]
        expected0 = df['-0bd'].tolist()
        expected1 = df['-1bd'].tolist()
        self.assertEqual(bd0,expected0)
        bd1 = df.apply(lambda x:handle_sample_data_dates(x,'1bd','-'),axis=1)
        bd1 = [l.strftime('%Y-%m-%d') for l in bd1.tolist()]
        self.assertEqual(bd1,expected1)
    
    def test_ddhCases(self):
        '''
            test edge cases of ddh and dates subtraction
        '''
        self.assertEqual(ddh('6m+6m'),durationModule.Duration(years=1, months=0, days=0, modified=False, debug=False))
        with self.assertRaises(Exception) as cm1:
            ddh('t+7dm')
        with self.assertRaises(Exception) as cm2:
            ddh('t+7g')
        
        self.assertEqual(ddh('6m+3m+1y+1/5/24'), dateModule.Date(2025,10,5))
        self.assertEqual(ddh('6m+3m+1y'),durationModule.Duration(years=1, months=9, days=0, modified=False, debug=False))
        self.assertEqual(ddh('2y-5m-4d'),durationModule.Duration(years=1, months=7, days=-4, modified=False, debug=False))
        self.assertEqual(dateModule.Date(2024,5,5)-dateModule.Date(2022,10,9),
                         durationModule.Duration(years=1, months=6, days=26, modified=False, _anchor_start=dateModule.Date(2022,10,9), _anchor_end=dateModule.Date(2024,5,5), _anchor_months=19, _anchor_days=-4, debug=False))
        d1 = ddh('10/9/22 - 5/5/24')
        d2 = dateModule.Date(2022,10,9) - dateModule.Date(2024,5,5)
        self.assertEqual(d1,d2)
        self.assertEqual(d2 + dateModule.Date(2024,5,5),dateModule.Date(2022,10,9))
        d3 = ddh('5/5/24 - 10/9/22')
        self.assertEqual(d3,
                         durationModule.Duration(years=1, months=6, days=26, modified=False, _anchor_start=dateModule.Date(2022,10,9), _anchor_end=dateModule.Date(2024,5,5), _anchor_months=19, _anchor_days=-4, debug=False))
        self.assertEqual(ddh('1/12/24+11m13d|NYuWE'),dateModule.Date(2024,12,26))

class TestUsageMore(unittest.TestCase):
    def tests_from_excel(self):
        #### TURN DEBUG ON OFF ####
        # ddh.debug = True
        ddh.debug = False
        #### TURN DEBUG ON OFF ####
        import pandas as pd
        
        df = pd.read_excel('tests/test_cases.xlsx')
        cols = ['from','ParserString','final','convention']
        df = df[cols]
        df = df.dropna()
        c = 0
        for idx,row in df.iterrows():
            s = row['ParserString']
            f = row['from']
            convention = row['convention']
            _a = time.time()
            a = ddh(row['final'])
            b = ddh(s)
            _b = time.time()
            ans = f'{str(b==a):<5}'
            res = color(ans,'green') if b==a else color(ans,'red')
            ms = round((_b -_a)*1000,0)
            if ms > 100:
                ms = f'{ms:>10} ms'
                ms = color(ms,'red')
            else:
                ms = f'{ms:>10} ms'
            
            print(f'tesing: {s:>28} = {b} == {a}',res,ms)
            if a==b:
                c+=1
        print(f'{c}/{len(df)} passed')
        # assert True if an only if all 32 pass
        self.assertTrue(c==len(df))
    
    

if __name__=='__main__':  

    # x = ddh('20231231+0bd|NYuWE/MOD',convention='YMD')
    # print(x)
    unittest.main()



import re

# General patterns
OPERATOR = r" ?(\+|-)? ?"
INT_PART = r"\d+"
OPTIONAL_DECIMAL = r"(?:\.|\.\d+)?"
NUMBER = f"{INT_PART}{OPTIONAL_DECIMAL}"

# DateString Patterns (rough patterns, parser does validation)
D = M = r"\d{1,2}"
Y = r"\d{2,4}"
_ = r"(?: ?-|/|\A\Z)?"
YMD = f"{Y}{_}{M}{_}{D}"
MDY = f"{M}{_}{D}{_}{Y}"
DMY = f"{D}{_}{M}{_}{Y}"

# DurationString patterns
PERIOD_LETTER = r"(?:bd|BD|[dDwWmMqQsShHyY])"
DATE_PERIOD = f"(?:({NUMBER})({PERIOD_LETTER}))"
REPEATING_DATE_PERIODS = (
    f"(?:{DATE_PERIOD}{DATE_PERIOD}?{DATE_PERIOD}?{DATE_PERIOD}?{DATE_PERIOD}?)"
)
PIPE = r"(\|)?"
CAL = r"([A-Z]{2,3})"
OPTIONAL_CALS= f'(?:u{CAL})?'*7
CALS = f'{CAL}{OPTIONAL_CALS}'
REPEATING_CALUNIONS = f" ?(?:{PIPE}) ?{CALS}"
ROLL = r" ?(/)? ?(MOD) ?"
PIPE_ROLL = f"(?:{ROLL})?"
PIPE_REPEAT_CAL_UNION = f"(?:{REPEATING_CALUNIONS})?"
COMPLETE_DURATION = (
    f"({OPERATOR}{REPEATING_DATE_PERIODS} ?{PIPE_REPEAT_CAL_UNION}{PIPE_ROLL})"
)

# Compiled
YMD = re.compile(YMD)
MDY = re.compile(MDY)
DMY = re.compile(DMY)
COMPLETE_DURATION = re.compile(COMPLETE_DURATION)





# Next iteration of patterns
# DurationString patterns
# OPERATOR = r'([+-])'
# NUMBER = r'(\d+)'
# UNIT = r"(bd|BD|[dDwWmMqQsShHyY])"
# PERIOD = f'{NUMBER}{UNIT}'
# MULTIPLE_PERIODS = f'{PERIOD}'*4
# MULTIPLE_PERIODS_WITH_OPERATOR = f'{OPERATOR}? ?{MULTIPLE_PERIODS} ?'
# CAL = r"([A-Z]{2,3})"
# UNIONED_CALS = f'u{CAL}?'*7
# PIPE_WITH_CALS= CAL + f'(?:| ?{CAL}{UNIONED_CALS} ?) ?'
# MOD = r'(\MOD)?'
# DURATION_STRING = f'{MULTIPLE_PERIODS_WITH_OPERATOR} ?{PIPE_WITH_CALS}? {MOD} '



from setuptools import find_packages, setup

setup(
    name="dateroll",
    version="0.0.41",
    description="""dateroll makes working with dates less painful.""",
    author="Anthony Malizzio",
    author_email="anthony.malizzio@disent.com",
    license="Apache Software License",
    url="https://github.com/disentcorp/dateroll",
    classifiers=[
        "Development Status :: 3 - Alpha",
        "Intended Audience :: Science/Research",
        "Intended Audience :: Developers",
        "Programming Language :: Python",
        "Programming Language :: Python :: 3",
        "Programming Language :: Python :: 3.7",
        "Programming Language :: Python :: 3.8",
        "Programming Language :: Python :: 3.9",
        "Programming Language :: Python :: 3.10",
        "Programming Language :: Python :: 3.11",
        "Programming Language :: Python :: 3.12",
        "Programming Language :: Python :: 3 :: Only",
        "Topic :: Software Development",
        "Topic :: Scientific/Engineering",
        "License :: OSI Approved :: Apache Software License",
        "Topic :: Utilities",
        "Operating System :: Unix",
    ],
    packages=[
        "dateroll",
        "dateroll.ddh",
        "dateroll.calendars",
        "dateroll.duration",
        "dateroll.parser",
        "dateroll.schedule",
        "dateroll.date",
    ],
    include_package_data=True,
    package_data={"dateroll": ["sampledata/*.csv", "tests/*"]},
    install_requires=["python-dateutil"],
    license_files=("LICENSE",),
    python_requires=">=3.7",
    zip_safe=False,
)




import code
from dateroll.calendars.calendarmath import calmath
from dateroll import pretty

class Schedule(list): 
    def __init__(self,start,stop,step,origin_string=None):
        self.start = start
        self.stop = stop
        self.step = step
        self.origin_string = origin_string
        if self.step >=0:
            self.direction = 'forward'
        else:
            self.direction = 'backward'

        self.cals = step.cals

        self.run()
        self.num_dates = len(self.dates)

    def __len__(self):
        return self.num_dates

    def run(self):
        '''
            gives the date range
            direction of date generations is given by the sign of step
        '''
        dates = []

        # backward generation
        if self.direction == 'forward':
            cursor = self.start
            while cursor < self.stop:
                dates.append(cursor)
                # we use plus sign because step<0
                cursor += self.step
            dates.append(self.stop)
        else:
            # backward generation
            cursor = self.stop
            while cursor > self.start:
                dates.append(cursor)
                cursor -= -self.step
            dates.append(self.start)


        self.dates = sorted(dates)
    
    @property
    def cal(self):
        _ = pretty.pretty_between_two_dates(self.start,self.stop,self.cals,calmath)
        print(_)
    
    def __str__(self):
        s = f"""Schedule:
    start      : {self.start}
    stop       : {self.start}
    step       : {self.start}
    direction  : {self.direction}
    cals       : {self.cals}
    orig str   : {self.origin_string}
    #dates     : {self.num_dates}"""
        return s
    
    
    def __repr__(self):
        constructor = ""
        for k, v in self.__dict__.items():
            if k not in ['run','spit','dates','debug'] and v:
                constructor += f"{k}={str(v)}, "
        return f'{self.__class__.__name__}({constructor.rstrip(", ")})'
    
    @property
    def split(self):
        import pandas as pd
        list_of_dates = self.dates
        start = list_of_dates[:-1]
        stop = list_of_dates[1:]
        step = [self.step.to_string()]*(len(list_of_dates)-1)
        df = pd.DataFrame({'start':start,'stop':stop,'step':step})
        df.index.name = 'per'
        return df
    
    @property
    def split_bond(self):
        '''
        makes a bond schedule assuming T+0BD on weekends
        '''
        df = self.split
        df['type']='interest'
        df.columns = ['starts','ends','days','type']
        df['pays']=df['ends']+'0bd'
        st,ed = min(self.dates),max(self.dates)
        
        df.loc[-1]=[None,None,None,'principal',st+'0bd']
        df.loc[len(df)-1]=[None,None,None,'repayment',ed+'0bd']
        df.index+=1
        df['days']=df.apply(lambda row: (row["ends"]-row['starts']).just_exact_days if row['starts'] else None,axis=1)
        return df.sort_values(by='pays')

    def to_string(self):
        '''
        should print as string
        '''
        if self.origin_string:
            return self.origin_string
        else:
            '''
            should subtract from t for a today's version of the string
            '''
            raise NotImplementedError


import unittest
import datetime

from dateroll.schedule.schedule import Schedule
from dateroll.date.date import Date
from dateroll.duration.duration import Duration

class TestSchedule(unittest.TestCase):
    @classmethod
    def setUpClass(cls): ...

    @classmethod
    def tearDownClass(self): ...

    def test_dates_forward(self):
        '''
            test dates which give a range of dates and step>0
        '''
        
        start = Date(2023,1,1)
        stop = Date(2023,2,1)
        step = Duration(bd=1)
        sch = Schedule(start,stop,step)
        dts = sch.dates
        # convert into datetime
        dts = [d.datetime for d in dts]
        expected = [datetime.datetime(2023, 1, 1, 0, 0), datetime.datetime(2023, 1, 2, 0, 0), datetime.datetime(2023, 1, 3, 0, 0), datetime.datetime(2023, 1, 4, 0, 0), datetime.datetime(2023, 1, 5, 0, 0), datetime.datetime(2023, 1, 6, 0, 0), datetime.datetime(2023, 1, 9, 0, 0), datetime.datetime(2023, 1, 10, 0, 0), datetime.datetime(2023, 1, 11, 0, 0), datetime.datetime(2023, 1, 12, 0, 0), datetime.datetime(2023, 1, 13, 0, 0), datetime.datetime(2023, 1, 16, 0, 0), datetime.datetime(2023, 1, 17, 0, 0), datetime.datetime(2023, 1, 18, 0, 0), datetime.datetime(2023, 1, 19, 0, 0), datetime.datetime(2023, 1, 20, 0, 0), datetime.datetime(2023, 1, 23, 0, 0), datetime.datetime(2023, 1, 24, 0, 0), datetime.datetime(2023, 1, 25, 0, 0), datetime.datetime(2023, 1, 26, 0, 0), datetime.datetime(2023, 1, 27, 0, 0), datetime.datetime(2023, 1, 30, 0, 0), datetime.datetime(2023, 1, 31, 0, 0), datetime.datetime(2023, 2, 1, 0, 0)]
        self.assertEqual(dts,expected)

    def test_dates_backward(self):
        '''
            test dates when step<0
        '''
        start = Date(2023,2,1)
        stop = Date(2023,1,1)
        step = Duration(bd=-1)
        sch = Schedule(start,stop,step)
        dts = sch.dates
        # convert into datetime
        dts = [d.datetime for d in dts]
        expected = [datetime.datetime(2023, 1, 2, 0, 0), datetime.datetime(2023, 1, 3, 0, 0), datetime.datetime(2023, 1, 4, 0, 0), datetime.datetime(2023, 1, 5, 0, 0), datetime.datetime(2023, 1, 6, 0, 0), datetime.datetime(2023, 1, 9, 0, 0), datetime.datetime(2023, 1, 10, 0, 0), datetime.datetime(2023, 1, 11, 0, 0), datetime.datetime(2023, 1, 12, 0, 0), datetime.datetime(2023, 1, 13, 0, 0), datetime.datetime(2023, 1, 16, 0, 0), datetime.datetime(2023, 1, 17, 0, 0), datetime.datetime(2023, 1, 18, 0, 0), datetime.datetime(2023, 1, 19, 0, 0), datetime.datetime(2023, 1, 20, 0, 0), datetime.datetime(2023, 1, 23, 0, 0), datetime.datetime(2023, 1, 24, 0, 0), datetime.datetime(2023, 1, 25, 0, 0), datetime.datetime(2023, 1, 26, 0, 0), datetime.datetime(2023, 1, 27, 0, 0), datetime.datetime(2023, 1, 30, 0, 0), datetime.datetime(2023, 1, 31, 0, 0), datetime.datetime(2023, 2, 1, 0, 0), datetime.datetime(2023, 2, 1, 0, 0)]
        self.assertEqual(dts,expected)



if __name__=='__main__':
    unittest.main()


import traceback
import datetime
import re
import code

# from dateroll.date.date import Date

import dateroll.date.date as dt
import dateroll.duration.duration as dur
from dateroll.parser import patterns
from dateroll.schedule.schedule import Schedule
from dateroll import utils
from dateroll.settings import settings
import dateroll.parser.parser as parserModule


TODAYSTRINGVALUES = ["today", "t0", "t"]


class ParserStringsError(Exception): ...


def parseTodayString(s):
    """
        this is [the] place where "t" is replaced
    """
    today = datetime.date.today()
    fmt = utils.convention_map[settings.convention]
    today_string = today.strftime(fmt)

    for t in TODAYSTRINGVALUES:
        # search order matters in TODAYSTRINGVALUES
        if t in s:
            return s.replace(t, today_string)
    return s


def parseDateString(s,gen):
    """
    for a given convention, see if string contains 1 or 2 dates
    regex to extract string, and Date.from_string(), which calls dateutil.parser.parse

    valid DateStrings for refence:

        MDY: american: 1 or 2 digit month, 1 or 2 digit day, and 2 or 4 digit year
        DMY: european: 1 or 2 digit day, 1 or 2 digit month, and 2 or 4 digit year
        YMD: international: 2 or 4 digit year, 1 or 2 digit month, 1 or 2 digit day

    """
    
    if settings.convention == "MDY":
        pattern = patterns.MDY
        dateparser_kwargs = {"dayfirst": False,"yearfirst": False}
    elif settings.convention == "DMY":
        pattern = patterns.DMY
        dateparser_kwargs = {"dayfirst": True, "yearfirst": True}
    elif settings.convention == "YMD":
        pattern = patterns.YMD
        dateparser_kwargs = {"yearfirst": True,"dayfirst":False}

    dates = {}
    matches = re.findall(pattern, s)
    res = s
    for match in matches:
        date = dt.Date.from_string(match, **dateparser_kwargs)
        next_letter = next(gen())
        # match only 1st time!! or causes letter mismatch
        res = res.replace(match, '+'+next_letter,1)
        dates[next_letter]=date
    
    return dates, res


def process_duration_match(m: tuple):
    """
    COMPLETE_DURATION regex matches a 23-item tuple
    This function extracts the parts and calls the Duration contructor appropriately
    Starts with empty, and adds as finds from incoming tuple
    """
    # operator becomes multiplier
    duration_contructor_args = {}

    # get initial multplier (if any)
    op = m[1]
    if op == "+" or op == "":
        mult = 1
    elif op == "-":
        mult = -1
    else:
        raise Exception("Unknown operator")

    # get all the pairs
    for i in range(2, 12, 2):
        number = m[i]
        unit = m[i + 1]

        if number and unit:
            # cast number to integer y,s,q,m,w,d
            number = int(number)
            if i < 4:
                # use multiplier on first pair
                if unit=='bd':
                    
                    number = float(number) * mult
                else:
                    number *= mult

            duration_contructor_args[unit] = number
    
    # attach calendars if any
    cals = m[13:21]
    
    # remove empty strings from cals
    cals = tuple(filter(lambda x:x!='',cals))
    for cal in cals:
        (
            duration_contructor_args.setdefault("cals", []).append(cal)
            if cal and cal.isupper()
            else None
        )

    # attach roll if any
    modified = m[22]
    if modified:
        duration_contructor_args["modified"] = True

    # add signed bd's if implicit scenario
    if 'bd' not in duration_contructor_args and 'BD' not in duration_contructor_args:
        if modified or len(cals) > 0:
            duration_contructor_args['bd'] = float(0.0)*mult
    
    duration = dur.Duration(**duration_contructor_args)
    return duration

def parseCalendarUnionString(s):
    '''
    use regex for validation of calendar string, if valid, then split on 'u' is SAFE
    '''
    matches = re.match(patterns.CALS,s)
    if matches:
        cals = tuple(sorted(s.split('u')))
        return cals
    else:
        raise Exception(f'{s} not a valid calendar string')

def parseDurationString(s,gen):
    """
    check for any DurationString:

        units:      1-9
        period's:   d,D day
                    bd,BD business day
                    w,W week
                    m,M month
                    q,Q quarter
                    y,Y year
        e.g. 1d, or 3M, or 9Y

        they can repeat: 1y3m9d = 1y + 3m + 9d

        calendars after |
            calendar as any uppercase 2 letter combo that map's to installed calendars
                WE -> by default is weekend calendar (list of all sat and sun from -100y to +100y)
                NY -> new york federal holidays
                EU -> ECB holidays
            calender unions: repreating calendars with "u" for union
                WEuNY -> all weekend holidays + NY holidays
                WUuNYuEU -> union of all 3 sets
        modifier after  /
                /MOD means modified the direction of travel for bd's to stay in current month
    """
    durations = {}
    matches = re.findall(patterns.COMPLETE_DURATION, s)

    for m in matches:
        duration = process_duration_match(m)
        dur_str = m[0]
        
        next_letter = next(gen())
        s = s.replace(dur_str,'+'+next_letter,1)
        durations[next_letter] = duration
    
    return durations, s


def parseDateMathString(s, things):
    """
    Looks for any linear formula with plus or minus
    Uses substitution and evaluation. Safe because wouldn't get here if *things were not validated, and regex didn't match.
    """
    before = s
    math_matches = re.match(r'[A-Z]', s)

    # we use this validation here to make the len(things)==1 condition work
    letters_used = re.findall(r'[A-Z]',s)
     
    # bad case, letter mismatch
    for i in letters_used:
        if i not in things:
            print(s,things)
            raise ParserStringsError("Cannot recognize as date math", s)
    for j in things:
        if j.isalpha():
            if j not in letters_used:
                print(s,things)
                raise ParserStringsError("Cannot recognize as date math", s)
        
    # good case, do the math
    gs = {k: v for k, v in things.items()}
            
    try:
        total = eval(s,{},gs)
        return total
    except Exception as e:
        raise ParserStringsError("Cannot recognize as date math", s)


def parseScheduleString(s):  # pragma: no cover
    """ """
    raise NotImplementedError

if __name__=='__main__':  # pragma: no cover
    s = ' A+B'
    s2 = ' +A+B -C'
    s3 = ' +A'
    things = {'A':4,'B':5}
    things2 = {'A':4}
    rs = parseDateMathString(s,things)
    print(rs)
     